import functools
import inspect
from typing import Any, Callable, TypeVar
import warnings

T = TypeVar("T")
LAZY_PREFIX = "_pgmlazy_"


class RefreshProperty(object):
    """An instance of this class marks a lazy-evaluated property as requiring a refresh"""

    def __eq__(self, __o: object) -> bool:
        return self.__class__ == __o.__class__


def get_lazy_property(obj: object, name: str, func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    """Retrieves a lazy property value"""
    attribute = LAZY_PREFIX + name
    has_attribute = hasattr(obj, attribute)
    if not has_attribute or (has_attribute and getattr(obj, attribute) == RefreshProperty()):
        setattr(obj, attribute, func(*args, **kwargs))
    return getattr(obj, attribute)


def lazy_property(func):
    # not used for now because of https://github.com/microsoft/pylance-release/discussions/2716
    # get_lazy_property implements that logic without having to use a decorator

    @property
    @functools.wraps(func)
    def _wrapper(self):
        return get_lazy_property(self, func.__name__, func, self)

    return _wrapper


class DeprecatedWarning(UserWarning):
    pass


def deprecated(instructions):
    """Flags a method as deprecated.

    Args:
        instructions: A warning displayed to a user. E.g. 'Use my_func() instead.'
    """

    def decorator(func):
        """This is a decorator which can be used to mark functions
        as deprecated. It will result in a warning being emitted
        when the function is used."""

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            message = "Call to deprecated function {}. {}".format(func.__name__, instructions)

            frame = inspect.currentframe().f_back

            warnings.warn_explicit(
                message,
                category=DeprecatedWarning,
                filename=inspect.getfile(frame.f_code),
                lineno=frame.f_lineno,
            )

            return func(*args, **kwargs)

        return wrapper

    return decorator



"""PGMob will try to abstract OS commands using classes that provide support for a specific OS."""
from abc import abstractmethod
import shlex
from . import util
from .errors import PostgresShellCommandError


class _BaseShellEnv(object):
    """Base OS class interface that outlines necessary shell-dependent operations"""

    @staticmethod
    @abstractmethod
    def quote(cmd: str) -> str:
        """Quotes command line argument for a shell function
        Args:
            cmd (str): command line argument to escape

        Returns:
            str: quoted command line argument
        """
        raise NotImplementedError()

    @staticmethod
    def join_path(*args) -> str:
        """Joins path according to the local OS rules. Deals with URLs and Posix paths

        Args:
            *args: paths that need to be joined

        Returns:
            str: Joined path string
        """
        path = ""
        for arg in args:
            path = path + ("/" if path and arg else "") + (arg.strip("/") if path else arg.rstrip("/"))
        return path

    @staticmethod
    @abstractmethod
    def get_os_command_wrapper() -> str:
        """Returns a wrapper for an OS command used by run_os_command method

        Returns:
            str: wrapper for an OS command
        """
        raise NotImplementedError()


class ShellEnv(_BaseShellEnv):
    """Shell-like command line operations (/bin/sh)"""

    @staticmethod
    def quote(cmd: str) -> str:
        """Quotes command line argument for a shell function

        Args:
            cmd (str): command line argument to escape

        Returns:
            str: quoted command line argument"""
        return shlex.quote(cmd)

    @staticmethod
    def get_os_command_wrapper() -> str:
        return util.get_shell("run_postgres_command")


class OSCommandResult(object):
    """Results of the OS command execution

    Attributes:
        exit_code (int): command exit code
        text (str): stdout and stderr together
    """

    _exit_code: int
    text: str

    def __init__(self, command: str) -> None:
        self.command = command

    @property
    def exit_code(self) -> int:
        return self._exit_code

    @exit_code.setter
    def exit_code(self, value: int):
        try:
            self._exit_code = int(value)
        except (ValueError, TypeError):
            raise PostgresShellCommandError(self.command, self.text)

    def raise_for_error(self):
        """Raises an exception when exist code is not 0.

        Raises:
            PostgresShellCommandError: when exit code is not 0
        """
        if self.exit_code != 0:
            raise PostgresShellCommandError(self.command, self.text)



"""Backup and Restore tools

In order to execute a backup or a restore operation, PGMob would issue a ``pg_dump``/``pg_restore`` command
on a remote server spawned as a subprocess of the PostgreSQL process. The binary should be available on the
server (you can specify an exact path) and should be able to connect to localhost without a password.

Example:
    Create a backup from one database and restore it as a different database

    >>> old_db = "foo"  # doctest: +SKIP
    >>> new_db = "bar"  # doctest: +SKIP
    >>> file_backup = FileBackup(cluster=cluster)
    >>> file_backup.backup(database=old_db, path="/tmp/db.bak")
    >>> file_restore = FileRestore(cluster=cluster)
    >>> file_restore.restore(database=new_db, path="/tmp/db.bak")
"""
from typing import List, Optional
import logging
from pathlib import Path
from .os import ShellEnv, _BaseShellEnv
from . import cluster

LOGGER = logging.getLogger(__name__)


class _CommonOptions(object):
    """Common backup/restore options"""

    def __init__(self, shell: Optional[_BaseShellEnv] = None, **kwargs) -> None:
        self.shell = shell if shell else ShellEnv()
        self._add_if_exists = False
        self._clean = False
        self._create = False
        self._data_only = False
        self._exclude_schemas: List[str] = []
        self._format: Optional[str] = None
        self._no_privileges = False
        self._no_publications = False
        self._no_subscriptions = False
        self._no_tablespaces = False
        self._no_owner = False
        self._schema_only = False
        self._schemas: List[str] = []
        self._section: Optional[str] = None
        self._set_role: Optional[str] = None
        self._strict_names = False
        self._superuser: Optional[str] = None
        self._tables: List[str] = []
        self._verbose = False

        for k, v in kwargs.items():
            if getattr(self, "_" + k):
                setattr(self, "_" + k, v)
            else:
                raise AttributeError("Unknown attribute: %s", k)

    # enforcing propert data types

    @property
    def add_if_exists(self) -> bool:
        return self._add_if_exists

    @add_if_exists.setter
    def add_if_exists(self, value: bool) -> None:
        self._add_if_exists = value

    @property
    def clean(self) -> bool:
        return self._clean

    @clean.setter
    def clean(self, value: bool) -> None:
        self._clean = bool(value)

    @property
    def create(self) -> bool:
        return self._create

    @create.setter
    def create(self, value: bool) -> None:
        self._create = bool(value)

    @property
    def data_only(self) -> bool:
        return self._data_only

    @data_only.setter
    def data_only(self, value: bool) -> None:
        self._data_only = bool(value)

    @property
    def exclude_schemas(self) -> List[str]:
        return self._exclude_schemas

    @exclude_schemas.setter
    def exclude_schemas(self, value: List[str]) -> None:
        self._exclude_schemas = [str(x) for x in value]

    @property
    def format(self) -> Optional[str]:
        return self._format

    @format.setter
    def format(self, value: Optional[str]) -> None:
        self._format = str(value) if value else None

    @property
    def no_privileges(self) -> bool:
        return self._no_privileges

    @no_privileges.setter
    def no_privileges(self, value: bool) -> None:
        self._no_privileges = bool(value)

    @property
    def no_publications(self) -> bool:
        return self._no_publications

    @no_publications.setter
    def no_publications(self, value: bool) -> None:
        self._no_publications = bool(value)

    @property
    def no_subscriptions(self) -> bool:
        return self._no_subscriptions

    @no_subscriptions.setter
    def no_subscriptions(self, value: bool) -> None:
        self._no_subscriptions = bool(value)

    @property
    def no_tablespaces(self) -> bool:
        return self._no_tablespaces

    @no_tablespaces.setter
    def no_tablespaces(self, value: bool) -> None:
        self._no_tablespaces = bool(value)

    @property
    def no_owner(self) -> bool:
        return self._no_owner

    @no_owner.setter
    def no_owner(self, value: bool) -> None:
        self._no_owner = bool(value)

    @property
    def schema_only(self) -> bool:
        return self._schema_only

    @schema_only.setter
    def schema_only(self, value: bool) -> None:
        self._schema_only = bool(value)

    @property
    def schemas(self) -> List[str]:
        return self._schemas

    @schemas.setter
    def schemas(self, value: List[str]) -> None:
        self._schemas = [str(x) for x in value]

    @property
    def section(self) -> Optional[str]:
        return self._section

    @section.setter
    def section(self, value: Optional[str]) -> None:
        self._section = str(value) if value else None

    @property
    def set_role(self) -> Optional[str]:
        return self._set_role

    @set_role.setter
    def set_role(self, value: Optional[str]) -> None:
        self._set_role = str(value) if value else None

    @property
    def strict_names(self) -> bool:
        return self._strict_names

    @strict_names.setter
    def strict_names(self, value: bool) -> None:
        self._strict_names = bool(value)

    @property
    def superuser(self) -> Optional[str]:
        return self._superuser

    @superuser.setter
    def superuser(self, value: Optional[str]) -> None:
        self._superuser = str(value) if value else None

    @property
    def tables(self) -> List[str]:
        return self._tables

    @tables.setter
    def tables(self, value: List[str]) -> None:
        self._tables = [str(x) for x in value]

    @property
    def verbose(self) -> bool:
        return self._verbose

    @verbose.setter
    def verbose(self, value: bool) -> None:
        self._verbose = bool(value)

    def render_args(self) -> List[str]:
        """Renders options as command line arguments

        Returns:
            List[str]: list of command line arguments
        """
        options: List[str] = []
        if self.clean:
            options.append("--clean")
        if self.create:
            options.append("--create")
        if self.data_only:
            options.append("--data-only")
        if self.schema_only:
            options.append("--schema-only")
        if self.superuser:
            options.append(f"--superuser={self.shell.quote(self.superuser)}")
        for table in self.tables:
            options.append(f'--table="{self.shell.quote(table)}"')
        for schema in self.schemas:
            options.append(f'--schema="{self.shell.quote(schema)}"')
        for schema in self.exclude_schemas:
            options.append(f'--exclude-schema="{self.shell.quote(schema)}"')
        if self.no_privileges:
            options.append("--no-privileges")
        if self.no_subscriptions:
            options.append("--no-subscriptions")
        if self.no_publications:
            options.append("--no-publications")
        if self.no_tablespaces:
            options.append("--no-tablespaces")
        if self.format:
            options.append(f"--format={self.shell.quote(self.format)}")
        if self.set_role:
            options.append(f'--role="{self.shell.quote(self.set_role)}"')
        if self.add_if_exists:
            options.append("--if-exists")
        if self.strict_names:
            options.append("--strict-names")
        if self.verbose:
            options.append("--verbose")
        if self.section:
            options.append(f"--section={self.shell.quote(self.section)}")
        if self.no_owner:
            options.append("--no-owner")
        return options


class BackupOptions(_CommonOptions):
    """Parameters to be used for the pg_dump binary

    Class attributes match closely to those displayed by the "pg_dump --help" command.

    Args:
        shell (_BaseShellEnv):            shell processor that defines pathing and escaping for the current environment
        **kwargs (dict):                  any supporter attribute can be submitted as kwargs

    Attributes:
        shell (_BaseShellEnv):            shell processor that defines pathing and escaping for the current environment
        format (str):                     (c|d|t|p) output file format (custom(default), directory, tar, plain text)
        jobs (int):                       use this many parallel jobs to dump
        verbose (bool):                   verbose mode
        compress (int):                   (0-9) compression level for compressed formats
        lock_wait_timeout (int):          fail after waiting TIMEOUT for a table lock
        data_only (bool):                 dump only the data, not the schema
        clean (bool):                     clean (drop) database objects before recreating
        create (bool):                    include commands to create database in dump
        schemas (List[str]):              dump the named schemas only
        exclude_schemas (List[str]):      do NOT dump the named schemas
        no_owner (bool):                  skip restoration of object ownership in plain-text format
        schema_only (bool):               dump only the schema, no data
        superuser (str):                  superuser user name to use in plain-text format
        tables (List[str]):               dump the named tables only
        exclude_tables (List[str]):       do NOT dump the named tables
        no_privileges (bool):             do not dump privileges (grant/revoke)
        exclude_table_data (List[str]):   do NOT dump data for the named table(s)
        add_if_exists (bool):             use IF EXISTS when dropping objects
        as_inserts (bool):                dump data as INSERT commands, rather than COPY
        no_subscriptions (bool):          do not dump subscriptions
        no_publications (bool):           do not dump publications
        no_tablespaces (bool):            do not dump tablespace assignments
        section (str):                    dump named section (pre-data, data, or post-data)
        strict_names (bool):              require table and/or schema include patterns to match at least one entity each
        set_role (str):                   invoke SET ROLE before dump
        blobs (bool):                     Include large objects in the dump
    """

    def __init__(self, shell: Optional[_BaseShellEnv] = None, **kwargs) -> None:
        self._compress = False
        self._compression_level = 5
        self._exclude_tables: List[str] = []
        self._exclude_table_data: List[str] = []
        self._as_inserts = False
        self._create_database = False
        self._lock_wait_timeout: Optional[int] = None
        self._blobs: Optional[bool] = None
        super().__init__(shell=shell, **kwargs)
        self.format = "c"

    @property
    def compress(self) -> bool:
        return self._compress

    @compress.setter
    def compress(self, value: bool) -> None:
        self._compress = bool(value)

    @property
    def as_inserts(self) -> bool:
        return self._as_inserts

    @as_inserts.setter
    def as_inserts(self, value: bool) -> None:
        self._as_inserts = bool(value)

    @property
    def create_database(self) -> bool:
        return self._create_database

    @create_database.setter
    def create_database(self, value: bool) -> None:
        self._create_database = bool(value)

    @property
    def compression_level(self) -> int:
        return self._compression_level

    @compression_level.setter
    def compression_level(self, value: int) -> None:
        self._compression_level = int(value)

    @property
    def exclude_tables(self) -> List[str]:
        return self._exclude_tables

    @exclude_tables.setter
    def exclude_tables(self, value: List[str]) -> None:
        self._exclude_tables = [str(x) for x in value]

    @property
    def exclude_table_data(self) -> List[str]:
        return self._exclude_table_data

    @exclude_table_data.setter
    def exclude_table_data(self, value: List[str]) -> None:
        self._exclude_table_data = [str(x) for x in value]

    @property
    def lock_wait_timeout(self) -> Optional[int]:
        return self._lock_wait_timeout

    @lock_wait_timeout.setter
    def lock_wait_timeout(self, value: Optional[int]) -> None:
        self._lock_wait_timeout = None if value is None else int(value)

    @property
    def blobs(self) -> Optional[bool]:
        return self._blobs

    @blobs.setter
    def blobs(self, value: Optional[bool]) -> None:
        self._blobs = None if value is None else bool(value)

    def render_args(self):
        """Renders options as command line arguments

        Returns:
            List[str]: A list of command line arguments
        """

        options = super().render_args()
        if self.compress:
            options.append(f"--compress={int(self.compression_level)}")
        if self.blobs is not None:
            options.append("--blobs" if self.blobs else "--no-blobs")
        if self.lock_wait_timeout:
            options.append(f"--lock-wait-timeout={int(self.lock_wait_timeout)}")
        if self.as_inserts:
            options.append("--inserts")
        if self.create_database:
            options.append("--create")
        for table in self.exclude_tables:
            options.append(f'--exclude-table="{table}"')
        for table in self.exclude_table_data:
            options.append(f'--exclude-table-data="{table}"')
        return options


class RestoreOptions(_CommonOptions):
    """Parameters to be used for the pg_restore binary

    Class attributes match closely to those displayed by the "pg_restore --help" command.

    Args:
        shell (_BaseShellEnv):            shell processor that defines pathing and escaping for the current environment
        **kwargs (dict):                  any supporter attribute can be submitted as kwargs

    Attributes:
        shell (_BaseShellEnv):            shell processor that defines pathing and escaping for the current environment
        format (str):                     (c|d|t) backup file format (should be automatic)
        jobs (int):                       use this many parallel jobs to restore
        verbose (bool):                   verbose mode
        data_only (bool):                 restore only the data, no schema
        clean (bool):                     clean (drop) database objects before recreating
        create (bool):                    create the target database
        exit_on_error (bool):             exit on error, default is to continue
        indexes (List[str]):              restore named indexes
        schemas (List[str]):              restore only objects in these schemas
        exclude_schemas (List[str]):      do not restore objects in these schemas
        use_list (str):                   use table of contents from this file for selecting/ordering output
        no_owner (bool):                  skip restoration of object ownership
        functions (List[str]):            (NAME(args)) restore named functions
        schema_only (bool):               restore only the schema, no data
        superuser (str):                  superuser user name to use for disabling triggers
        tables (List[str]):               restore named relations (table, view, etc.)
        triggers (List[str]):             restore named triggers
        no_privileges (bool):             skip restoration of access privileges (grant/revoke)
        single_transaction (bool):        restore as a single transaction
        disable_triggers (bool):          disable triggers during data-only restore
        add_if_exists (bool):             use IF EXISTS when dropping objects
        no_subscriptions (bool):          do not restore subscriptions
        no_publications (bool):           do not restore publications
        no_tablespaces (bool):            do not restore tablespace assignments
        section (str):                    restore named section (pre-data, data, or post-data)
        no_data_for_failed_tables (bool): do not restore data of tables that could not be created
        strict_names (bool):              require table and/or schema include patterns to match at least one entity each
        set_role (str):                   invoke SET ROLE before dump
    """

    def __init__(self, shell: Optional[_BaseShellEnv] = None, **kwargs) -> None:
        self._exit_on_error = False
        self._indexes: List[str] = []
        self._functions: List[str] = []
        self._triggers: List[str] = []
        self._jobs: Optional[int] = None
        self._use_list: Optional[str] = None
        self._single_transaction = False
        self._disable_triggers = False
        self._no_data_for_failed_tables = False
        super().__init__(shell=shell, **kwargs)

    @property
    def exit_on_error(self) -> bool:
        return self._exit_on_error

    @exit_on_error.setter
    def exit_on_error(self, value: bool) -> None:
        self._exit_on_error = bool(value)

    @property
    def indexes(self) -> List[str]:
        return self._indexes

    @indexes.setter
    def indexes(self, value: List[str]) -> None:
        self._indexes = [str(x) for x in value]

    @property
    def functions(self) -> List[str]:
        return self._functions

    @functions.setter
    def functions(self, value: List[str]) -> None:
        self._functions = [str(x) for x in value]

    @property
    def triggers(self) -> List[str]:
        return self._triggers

    @triggers.setter
    def triggers(self, value: List[str]) -> None:
        self._triggers = [str(x) for x in value]

    @property
    def jobs(self) -> Optional[int]:
        return self._jobs

    @jobs.setter
    def jobs(self, value: Optional[int]) -> None:
        self._jobs = int(value) if value else None

    @property
    def use_list(self) -> Optional[str]:
        return self._use_list

    @use_list.setter
    def use_list(self, value: Optional[str]) -> None:
        self._use_list = str(value) if value else None

    @property
    def single_transaction(self) -> bool:
        return self._single_transaction

    @single_transaction.setter
    def single_transaction(self, value: bool) -> None:
        self._single_transaction = bool(value)

    @property
    def disable_triggers(self) -> bool:
        return self._disable_triggers

    @disable_triggers.setter
    def disable_triggers(self, value: bool) -> None:
        self._disable_triggers = bool(value)

    @property
    def no_data_for_failed_tables(self) -> bool:
        return self._no_data_for_failed_tables

    @no_data_for_failed_tables.setter
    def no_data_for_failed_tables(self, value: bool) -> None:
        self._no_data_for_failed_tables = bool(value)

    def render_args(self):
        """Renders options as command line arguments

        Returns:
            List[str]: A list of command line arguments
        """
        options = super().render_args()
        if self.exit_on_error:
            options.append("--exit-on-error")
        for index in self.indexes:
            options.append(f'--index="{index}"')
        for function in self.functions:
            options.append(f'--function="{function}"')
        for trigger in self.triggers:
            options.append(f'--trigger="{trigger}"')
        if self.jobs:
            options.append(f"--jobs={int(self.jobs)}")
        if self.use_list:
            options.append(f"--use-list={self.use_list}")
        if self.single_transaction:
            options.append("--single-transaction")
        if self.disable_triggers:
            options.append("--disable-triggers")
        if self.no_data_for_failed_tables:
            options.append("--no-data-for-failed-tables")
        return options


class _BackupRestoreOperation(object):
    """Base backup/restore operation class that implements binary execution.

    Args:
        cluster (cluster.Cluster): Postgres cluster object
        binary_path (str): Path to the binary
        command (str): OS shell command to execute
        base_path (str): base path for the backups
        options (_CommonOptions): Options object
        shell (_BaseShellEnv): shell processor that defines pathing and escaping for the current environment
    """

    def __init__(
        self,
        cluster: cluster.Cluster,
        binary_path: str,
        command: str,
        base_path: str,
        options: _CommonOptions,
        shell: Optional[_BaseShellEnv] = None,
    ) -> None:
        self.shell = shell if shell else ShellEnv()
        self.command = command
        self.binary = binary_path
        self.cluster = cluster
        self.base_path = base_path.rstrip("/")
        self.on_start_commands: List[str] = []
        self.on_finish_commands: List[str] = []
        self.options = options

    def _exec_commands(self, commands: List[str], **kwargs):
        for command in commands:
            formatted_command = command.format(**kwargs)
            LOGGER.debug(f"Running {self.__class__.__name__} command: {formatted_command}")
            self.cluster.run_os_command(command=formatted_command)

    def execute_command(self, database: str, path: str):
        """Executes backup/restore commands and replaces the formatted commands with actual values.

        Args:
            database(str): database name
            path(str): backup path

        Returns:
            str: stdin and stdout of executed command
        """
        full_path = self.shell.join_path(self.base_path, path)
        params = dict(
            database=database,
            path=full_path,
            options=" ".join(self.options.render_args()),
            binary=self.binary,
            filename=Path(full_path).name,
        )
        self._exec_commands(self.on_start_commands, **params)
        try:
            result = self._exec_commands([self.command], **params)
        finally:
            self._exec_commands(self.on_finish_commands, **params)
        return result


class FileBackup(_BackupRestoreOperation):
    """File Backup class that performs a Backup operation to a local filesystem.
    Executes pg_dump on a target postgres cluster spawned as a subprocess of Postgres server process.

    Args:
        cluster (cluster.Cluster): Postgres cluster object
        base_path (str): Base backup path. If specified, the backup path would be considered relative to it.
        binary_path (str): Path to the pg_dump binary. Use when a specific binary version is needed or the binary is not in PATH.
        options (BackupOptions): backup options represented by BackupOptions class
        shell (_BaseShellEnv): shell processor that defines pathing and escaping for the current environment

    Attributes:
        options (BackupOptions): backup options represented by the BackupOptions class
        cluster (cluster.Cluster): Postgres cluster object
        binary (str): Path to the pg_dump binary
        shell (_BaseShellEnv): shell processor that defines pathing and escaping for the current environment
        on_start_commands (List[str]): commands to execute prior to launching the backup
        command (str): main backup command
        on_finish_commands (List[str]): commands to execute after backup is completed or failed

    Example:
        Backup a database schema for tables "a" and "b"

        >>> db = "foo"  # doctest: +SKIP
        >>> backup = FileBackup(cluster=cluster, base_path="/tmp")
        >>> backup.options.schema_only = True
        >>> backup.options.exclude_tables = ["a", "b"]
        >>> backup.backup(database=db, path=f"{db}.bak")
    """

    def __init__(
        self,
        cluster: cluster.Cluster,
        base_path: str = "",
        binary_path: str = "pg_dump",
        options: Optional[BackupOptions] = None,
        shell: Optional[_BaseShellEnv] = None,
    ):
        super().__init__(
            cluster=cluster,
            binary_path=binary_path,
            command='{binary} {options} -d "{database}" > "{path}"',
            base_path=base_path,
            options=options if options else BackupOptions(),
            shell=shell,
        )

    def backup(self, database, path):
        """Backup a database to the specified path

        Args:
            database (str):    name of the database to backup
            path (str):        path (relative or absolute) to backup to

        Returns:
            str: pg_dump stdout and stderr output
        """
        return self.execute_command(database=database, path=path)


class GCPBackup(FileBackup):
    """GCP Backup class that uploads backups to a GCP bucket.
    Executes "pg_dump" on a target postgres cluster spawned as a subprocess of Postgres server process.
    Requires "gsutil" command available on the server host.
    To configure authentication, run "gcloud auth login" under postgres OS user.

    Args:
        cluster (cluster.Cluster): Postgres cluster object
        bucket (str): (Optional) GCP bucket name to work with. If specified, the backup path would
            be considered relative of the bucket path
        binary_path (str): Path to the pg_dump binary. Use when a specific binary version is needed or the binary is not in PATH.
        options (BackupOptions): backup options represented by BackupOptions class
        shell (_BaseShellEnv): shell processor that defines pathing and escaping for the current environment

    Attributes:
        options (BackupOptions): backup options represented by the BackupOptions class
        cluster (cluster.Cluster): Postgres cluster object
        binary: Path to the pg_dump binary.
        on_start_commands (List[str]): commands to execute prior to launching the backup
        command (str): main backup command
        on_finish_commands (List[str]): commands to execute after backup is completed or failed

    """

    def __init__(
        self,
        cluster: cluster.Cluster,
        bucket: str = "",
        binary_path: str = "pg_dump",
        options: Optional[BackupOptions] = None,
        shell: Optional[_BaseShellEnv] = None,
    ):
        super().__init__(
            cluster=cluster, binary_path=binary_path, base_path=bucket, options=options, shell=shell
        )
        self.command = '{binary} {options} -d "{database}" | gsutil cp - "{path}"'


class FileRestore(_BackupRestoreOperation):
    """Restore class that performs a Restore operation from a local filesystem.
    Executes pg_restore on a target postgres cluster spawned as a subprocess of Postgres server process.

    Args:
        cluster (cluster.Cluster): Postgres cluster object
        base_path (str): Base backup path. If specified, the backup path would be considered relative to it.
        binary_path (str): Path to the pg_restore binary. Use when a specific binary version is needed or the binary is not in PATH.
        options (RestoreOptions): restore options represented by the RestoreOptions class
        shell (_BaseShellEnv): shell processor that defines pathing and escaping for the current environment

    Attributes:
        options (RestoreOptions): restore options represented by the RestoreOptions class
        cluster: Postgres cluster object
        binary: Path to the pg_restore binary
        on_start_commands (List[str]): commands to execute prior to launching the restore
        command (str): main restore command
        on_finish_commands (List[str]): commands to execute after restore is completed or failed


    Example:
        Restore objects into database bar changing object ownership to "bar"

        >>> old_db = "foo"  # doctest: +SKIP
        >>> new_db = "bar"  # doctest: +SKIP
        >>> role = "ownerrole"  # doctest: +SKIP
        >>> backup = FileBackup(cluster=cluster, base_path="/tmp")
        >>> backup.backup(database=old_db, path=f"{old_db}.bak")
        >>> restore = FileRestore(cluster=cluster, base_path="/tmp")
        >>> restore.options.schema_only = True
        >>> restore.options.no_owner = True
        >>> restore.options.tables = ["a", "b"]
        >>> restore.options.set_role = role
        >>> restore.restore(database=new_db, path=f"{old_db}.bak")
    """

    def __init__(
        self,
        cluster: cluster.Cluster,
        base_path: str = "",
        binary_path: str = "pg_restore",
        options: RestoreOptions = None,
        shell: Optional[_BaseShellEnv] = None,
    ):
        super().__init__(
            cluster=cluster,
            binary_path=binary_path,
            command='{binary} {options} -d "{database}" "{path}"',
            base_path=base_path,
            options=options if options else RestoreOptions(),
            shell=shell,
        )

    def restore(self, database: str, path: str):
        """Restore a backup into the specified database
        Args:
            database (str):    name of the database to connect to
            path (str):        path (relative or absolute) to the backup file

        Returns:
            str: pg_restore full output
        """
        return self.execute_command(database=database, path=path)


class GCPRestore(FileRestore):
    """GCP Restore class that manages restores from GCP buckets.
    Executes "pg_restore" on a target cluster spawned as a subprocess of Postgres server process.
    Requires "gsutil" command available on the server host.
    To configure authentication, run "gcloud auth login" under postgres OS user.
    GCP bucket file is first copied to temp_path, then restored from disk. The file is removed once restore is finished.

    Args:
        cluster (cluster.Cluster): Postgres cluster object
        bucket (str): (Optional) GCP bucket name to work with. If specified, the backup path would
            be considered relative of the bucket path
        binary_path (str): Path to the pg_restore binary. Use when a specific binary version is needed or the binary is not in PATH.
        temp_path (str): Path to a temporary folder, to which the backup would be copied to.
            Direct restores from the bucket via pipe has proven to be unreliable.

    Attributes:
        options (RestoreOptions): restore options represented by RestoreOptions class
        cluster: Postgres cluster to execute the restore
        binary: Path to the pg_restore binary.
        on_start_commands (List[str]): commands to execute prior to launching the restore
        command (str): main restore command
        on_finish_commands (List[str]): commands to execute after restore is completed or failed
        temp_path (str): Path to a temporary folder
    """

    def __init__(
        self,
        cluster: cluster.Cluster,
        bucket: str = "",
        binary_path: str = "pg_restore",
        temp_path: str = "/tmp",
        options: Optional[RestoreOptions] = None,
        shell: Optional[_BaseShellEnv] = None,
    ):
        super().__init__(
            cluster=cluster, binary_path=binary_path, base_path=bucket, options=options, shell=shell
        )
        self.temp_path = Path(temp_path)
        self.on_start_commands = [f'gsutil cp "{{path}}" "{self.temp_path}/{{filename}}"']
        self.on_finish_commands = [f'rm -f "{self.temp_path}/{{filename}}"']
        self.command = f'{{binary}} {{options}} -d "{{database}}" "{self.temp_path}/{{filename}}"'



"""HBA rules as collection of strings that ignore whitespace on comparison"""
import collections
import re
from typing import Any, Iterable, List, Optional, Tuple, Union, TYPE_CHECKING

from ..adapters import ProgrammingError
from ..sql import SQL, Literal
from ..adapters.base import BaseCursor
from ..errors import *
from . import generic


if TYPE_CHECKING:
    from ..cluster import Cluster


class HBARule(str):
    """A record in pg_hba.conf file.
    Whitespace is considered equal upon object comparison.

    Args:
        str: A line from pg_hba.conf file
    """

    def __eq__(self, other):
        left = self.split()
        right = str(other).split()
        return left == right

    def __ne__(self, other):
        return not self.__eq__(other)

    def _get_field(self, name: str) -> Optional[str]:
        fields = self.fields
        field_map = {}
        auth_options: List[str] = []
        for i in range(len(fields)):
            if fields[i].startswith("#"):
                # anything after is a comment
                break
            if i == 0:
                field_map["type"] = fields[i]
                continue
            if i == 1:
                field_map["database"] = fields[i]
                continue
            if i == 2:
                field_map["user"] = fields[i]
                continue
            if i == 3:
                if field_map["type"] == "local":
                    field_map["auth_method"] = fields[i]
                else:
                    field_map["address"] = fields[i]
                continue
            if i == 4:
                if field_map["type"] == "local":
                    auth_options.append(fields[i])
                elif re.match("\\d{1,3}\.\\d{1,3}\.\\d{1,3}\\.\\d{1,3}", fields[i]):
                    field_map["mask"] = fields[i]
                else:
                    field_map["auth_method"] = fields[i]
                continue
            if i == 5:
                if field_map["type"] == "local":
                    auth_options.append(fields[i])
                elif "mask" in field_map:
                    field_map["auth_method"] = fields[i]
                else:
                    auth_options.append(fields[i])
                continue
            if i > 5:
                auth_options.append(fields[i])
        if name == "auth_options":
            return " ".join(auth_options)
        else:
            return field_map.get(name, None)

    @property
    def fields(self) -> List[str]:
        """A tuple of pg_hba.conf fields extracted from the record"""
        return self.split()

    @property
    def type(self) -> Optional[str]:
        """The type field"""
        return self._get_field("type")

    @property
    def database(self) -> Optional[str]:
        """The database field"""
        return self._get_field("database")

    @property
    def user(self) -> Optional[str]:
        """The user field"""
        return self._get_field("user")

    @property
    def address(self) -> Optional[str]:
        """The address field"""
        return self._get_field("address")

    @property
    def mask(self) -> Optional[str]:
        return self._get_field("mask")

    @property
    def auth_method(self) -> Optional[str]:
        return self._get_field("auth_method")

    @property
    def auth_options(self) -> Optional[str]:
        return self._get_field("auth_options")


class HBARuleCollection(collections.UserList[HBARule], generic._ClusterBound):
    """A collection of HBA rules in pg_hba.conf file

    Args:
        list: A list of lines from pg_hba or HBARule objects
        cluster (str): Postgres cluster object

    Attributes:
        cluster (str): Postgres cluster object
    """

    def __init__(self, cluster: "Cluster"):
        collections.UserList.__init__(self, [])
        generic._ClusterBound.__init__(self, cluster=cluster)
        if cluster:
            self.refresh()

    def refresh(self):
        """Resets any pending changes and retrieves objects from the cluster"""

        self.clear()

        def task(cursor: BaseCursor):
            cursor.execute(
                SQL(
                    """
                CREATE TEMPORARY TABLE pg_hba(
                id int GENERATED ALWAYS AS IDENTITY,
                lines text
            );
            """
                )
            )
            try:
                hba_file = cursor.scalar(
                    SQL("SELECT setting FROM pg_settings WHERE name = %s"), ("hba_file",)
                )
                cursor.execute(SQL("COPY pg_hba(lines) FROM %s"), (hba_file,))
                cursor.execute(SQL("SELECT lines FROM pg_hba ORDER BY id"))
                lines = cursor.fetchall()
            except ProgrammingError as e:
                raise PostgresError(e)
            finally:
                cursor.execute(SQL("DROP TABLE pg_hba;"))
            return [x[0] for x in lines]

        lines = self.cluster.execute_with_cursor(task)
        self.data.extend([HBARule(x) for x in lines])

    def __contains__(self, o: object) -> bool:
        return HBARule(o) in self.data

    def __iadd__(self, other: Iterable[HBARule]):
        self.data.extend([HBARule(r) for r in other])
        return self

    def __add__(self, other: Iterable[HBARule]):
        self.data.extend([HBARule(r) for r in other])
        return self

    def extend(self, item: Iterable[HBARule]):
        """Add multiple HBA rules to the collection

        Args:
            item(Iterable[Any]): An iterable of lines from pg_hba as strings or HBARule object
        """
        self.data.extend([HBARule(x) for x in item])

    def append(self, item: Union[str, HBARule]):
        """Append an HBA rule to the collection

        Args:
            item(Union[str, HBARule]): A line from pg_hba or HBARule object
        """
        self.data.append(HBARule(item))

    def remove(self, item: Union[str, HBARule]):
        """Remove an HBA rule from the collection

        Args:
            item(Union[str, HBARule]): A line from pg_hba or HBARule object
        """
        self.data.remove(HBARule(item))

    def index(self, item: Union[str, HBARule], *args) -> int:
        """Return a first index of a matching HBA rule from the collection

        Args:
            item(Union[str, HBARule]): A line from pg_hba or HBARule object
        """
        return self.data.index(HBARule(item), *args)

    def insert(self, index: int, item: Union[str, HBARule]):
        """Insert an HBA rule into the rule collection with a certain index

        Args:
            item(Union[str, HBARule]): A line from pg_hba or HBARule object
            index(int): Position to use
        """
        self.data.insert(index, HBARule(item))

    def alter(self):
        """Write HBA collection back into pg_hba.conf on the Postgres server."""

        def task(cursor):
            cursor.execute(
                SQL(
                    """CREATE TEMPORARY TABLE pg_hba(
                        id int GENERATED ALWAYS AS IDENTITY,
                        lines text
                    )"""
                )
            )
            try:
                hba_file = cursor.scalar(
                    SQL("SELECT setting FROM pg_settings WHERE name = %s"), ("hba_file",)
                )
                # load current lines and make a copy
                cursor.execute(SQL("COPY pg_hba(lines) FROM %s"), (hba_file,))
                cursor.execute(
                    SQL("COPY (SELECT lines FROM pg_hba ORDER BY id) TO %s"), (f"{hba_file}.bak.pgm",)
                )
                cursor.execute(SQL("DELETE FROM pg_hba"))
                # upload new rules and write them to the file
                cursor.execute(
                    SQL("INSERT INTO pg_hba(lines) VALUES {values}").format(
                        values=SQL(", ").join([SQL("({value})").format(value=Literal(x)) for x in self])
                    )
                )
                cursor.execute(SQL("COPY (SELECT lines FROM pg_hba ORDER BY id) TO %s"), (hba_file,))
            except ProgrammingError as e:
                raise PostgresError(e)
            finally:
                cursor.execute(SQL("DROP TABLE pg_hba"))

        self.cluster.execute_with_cursor(task)



"""Postgresql largeobject objects"""
from typing import TYPE_CHECKING, Optional
from pgmob.sql import SQL, Literal
from pgmob.adapters.base import BaseLargeObject
from pgmob.errors import *
from pgmob import util
from . import generic

if TYPE_CHECKING:
    from ..cluster import Cluster


class LargeObject(generic._DynamicObject, generic._CollectionChild):
    """
    Postgres LargeObject object. Represents a large object on a Postgres server.

    Args:
        oid (int): LargeObject OID
        cluster (str): Postgres cluster object
        definition (tuple): definition of the largeobject; used internally

    Attributes:
        oid (int): LargeObject OID
        cluster (str): Postgres cluster object
        owner (str): LargeObject owner
    """

    def __init__(
        self,
        oid: int = None,
        cluster: "Cluster" = None,
        parent: "LargeObjectCollection" = None,
        owner: str = None,
    ):
        """Initialize a new LargeObject object"""
        super().__init__(kind="LARGE OBJECT", cluster=cluster, oid=oid, name=str(oid))
        generic._CollectionChild.__init__(self, parent=parent)
        self._owner = owner

    def _sql_fqn(self) -> Literal:
        return Literal(self._oid)

    def _with_lobject(self, task, mode="rw"):
        cluster = self.cluster
        with cluster._no_autocommit():
            lo = cluster.adapter.lobject(self._oid, mode=mode)
            result = task(lo)
            lo.close()
            return result

    @property
    def owner(self) -> Optional[str]:
        return self._owner

    @owner.setter
    def owner(self, owner: str):
        generic._set_ephemeral_attr(self, "owner", owner)

    def drop(self):
        """Drops the largeobject from the Postgres cluster"""
        self._with_lobject(lambda lo: lo.unlink())

    def refresh(self):
        """Re-initializes the object, refreshing its properties"""
        super().refresh()
        if not self._ephemeral:
            sql = util.get_sql("get_large_object") + SQL(" WHERE lo.oid = %s")
            result = self.cluster.execute(sql, (self._oid,))
            if not result:
                raise PostgresError("LargeObject not found")
            mapper = _LargeObjectMapper(result[0])
            mapper.map(self)

    def write(self, data: bytes, truncate: bool = True):
        """Overwrites Large object contents

        Args:
            data(bytes): Large Object contents
            truncate(bool): Whether to truncate the object before write. Default = True
        """

        def task(lob: BaseLargeObject):
            try:
                if truncate:
                    lob.truncate()
                lob.write(data)
            finally:
                lob.close()

        self._with_lobject(task, mode="w")

    def read(self, mode="t"):
        """Read Large object contents

        Args:
            mode(str):
                - r  Open for read only
                - w  Open for write only
                - rw Open for read/write
                - n  Don’t open the file
                - b  Return data as bytes
                - t  Decode data as string

        Returns:
            bytes/str: Large object contents
        """
        return self._with_lobject(lambda lo: lo.read(), mode=mode)

    def truncate(self, len: int = 0):
        """Truncate Large object contents.

        Args:
            len(int): truncate to this number of bytes
        """
        self._with_lobject(lambda lo: lo.truncate(len), mode="w")


class _LargeObjectMapper(generic._BaseObjectMapper[LargeObject]):
    """Maps out a resultset from a database query to a large object"""

    attributes = [
        "oid",
        "owner",
    ]


class LargeObjectCollection(generic._BaseCollection[LargeObject]):
    """An iterable collection of largeobjects indexed by name."""

    def __init__(self, cluster):
        super().__init__(cluster=cluster)
        if cluster:
            self.refresh()

    def refresh(self):
        """Resets any pending changes and retrieves objects from the cluster"""

        super().refresh()
        sql = util.get_sql("get_large_object")
        result = self.cluster.execute(sql)
        for mapper in [_LargeObjectMapper(x) for x in result]:
            self[mapper["oid"]] = mapper.map(
                LargeObject(cluster=self.cluster, oid=mapper["oid"], parent=self)
            )



"""Schema objects. Represents schemas on the Postgres cluster"""
from typing import TYPE_CHECKING, Any, Dict, Optional, Union
from ..sql import SQL, Composable, Identifier
from ..errors import *
from .. import util
from . import generic


if TYPE_CHECKING:
    from ..cluster import Cluster


class Schema(generic._DynamicObject, generic._CollectionChild):
    """Postgres schema object. Represents a schema object on a Postgres cluster.

    Args:
        name (str): schema name
        owner (str): schema owner
        oid (int): Schema OID
        cluster (str): Postgres cluster object
        parent (SchemaCollection): parent collection

    Attributes:
        name (str): Schema name
        cluster (str): Postgres cluster object
        owner (str): Schema owner
        oid (int): Schema OID
    """

    def __init__(
        self,
        name: str,
        owner: str = None,
        cluster: "Cluster" = None,
        parent: "SchemaCollection" = None,
        oid: Optional[int] = None,
    ):
        super().__init__(cluster=cluster, name=name, kind="SCHEMA", oid=oid)
        generic._CollectionChild.__init__(self, parent=parent)
        self._owner = owner

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str):
        generic._set_ephemeral_attr(self, "name", name)

    @property
    def owner(self) -> Optional[str]:
        return self._owner

    @owner.setter
    def owner(self, owner: str):
        generic._set_ephemeral_attr(self, "owner", owner)

    def drop(self, cascade: bool = False):
        """Drops the schema from the Postgres cluster

        Args:
            cascade (bool): drop dependent objects
        """

        sql = SQL("DROP SCHEMA {schema}").format(schema=self._sql_fqn())
        if cascade:
            sql += SQL(" CASCADE")
        self.cluster.execute(sql)

    def refresh(self):
        """Re-initializes the object, refreshing its properties"""
        super().refresh()
        if not self._ephemeral:
            sql = util.get_sql("get_schema") + SQL(" WHERE n.oid = %s")
            result = self.cluster.execute(sql, (self.oid,))
            if not result:
                raise PostgresError("Schema %s with oid %s was not found", self.name, self.oid)
            mapper = _SchemaMapper(result[0])
            mapper.map(self)

    def create(self):
        """Create a database on the Postgres cluster"""
        self.cluster.execute(self.script(as_composable=True))
        sql = util.get_sql("get_schema") + SQL(" WHERE n.nspname = %s")
        _SchemaMapper(self.cluster.execute(sql, self.name)[0]).map(self)

    def script(self, as_composable: bool = False) -> Union[str, Composable]:
        """Generate a schema creation script.

        Args:
            as_composable (bool): return Composable object instead of plain text

        Returns:
            Union[str, Composable]: schema creation script
        """
        sql = "CREATE SCHEMA {schema}"
        params: Dict[str, Composable] = {"schema": self._sql_fqn()}
        if self.owner:
            sql += " AUTHORIZATION {owner}"
            params["owner"] = Identifier(self.owner)

        command = SQL(sql).format(**params)
        if as_composable:
            return command
        else:
            with self.cluster.adapter.cursor() as cur:
                return cur.mogrify(command)


class _SchemaMapper(generic._BaseObjectMapper[Schema]):
    """Maps out a resultset from a database query to a schema object"""

    attributes = [
        "name",
        "owner",
        "oid",
    ]


class SchemaCollection(generic._BaseCollection[Schema]):
    """An iterable collection of schemas indexed by name."""

    def __init__(self, cluster: "Cluster"):
        super().__init__(cluster=cluster)
        if cluster:
            self.refresh()

    def refresh(self):
        """Resets any pending changes and retrieves child objects from the cluster"""
        super().refresh()
        sql = util.get_sql("get_schema")
        result = self.cluster.execute(sql)
        for mapper in [_SchemaMapper(x) for x in result]:
            self[mapper["name"]] = mapper.map(Schema(cluster=self.cluster, name=mapper["name"], parent=self))

    def new(
        self,
        name: str,
        owner: str = None,
    ) -> Schema:
        """Create a schema object on the current Postgres cluster. The object
        is created ephemeral and either needs to be added to the schema collection,
        or .create() needs to be executed.

        Args:
            name (str): The name of the database
            owner (str): Owner of the database
        """
        return Schema(
            cluster=self.cluster,
            name=name,
            parent=self,
            owner=owner,
        )

    def add(self, schema: Schema):
        """Adds the schema object to the collection, simultaneously creating it on the cluster

        Args:
            schema (Schema): initialized schema object
        """
        schema.cluster = self.cluster
        schema._parent = self
        schema.create()
        self[schema.name] = schema



"""PGMob uses internal classes that wrap plaintext SQL queries and ensure safe parameter validation. These
classes are eventually decoded by adapters into SQL statements with appropriate syntax, parameters and quotation.
"""

import string

from typing import Generator, List, Sequence, Union


class Composable:
    """Common interface for SQL-like objects"""

    def __add__(self, other: "Composable") -> "Composed":
        return Composed(self, other)

    def __mul__(self, other: int) -> "Composed":
        if not isinstance(other, int):
            raise TypeError("int type is required")
        return Composed(*[self for _ in range(other)])

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Composable):
            return self.compose() == other.compose()
        else:
            return False

    def compose(self) -> "Composed":
        """Method utilized by an adapter to retrieve a Composed object that contains
        a list of objects and is ready to be iterated upon. All iterated objects are
        guaranteed to be one of: SQL, Literal, Identifier.

        Returns:
            Composed: a composed iterable object
        """
        return Composed(self)


class _Singleton(Composable):
    """A single SQL query part"""

    def __init__(self, value: object) -> None:
        self._value = value

    def __str__(self) -> str:
        return self.__class__.__name__ + f'("{str(self._value)}")'

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Composed):
            return super().__eq__(other)
        return isinstance(other, self.__class__) and self._value == other._value

    def value(self) -> object:
        """Retrieve the underlying value of a Composable: a query string, identifier name
        or a value of a literal. Not supported on Composed objects, which need to be
        iterated upon first before calling .value() on the yielded Composable objects.

        Returns:
            object: the underlying object in a Composable
        """
        return self._value


class Composed(Composable):
    """Composed object. Finalized state of co-joined query parts. Can be
    iterated upon to retrieve the individual parts in proper order."""

    def __init__(self, *args: Composable) -> None:
        self._parts = list(self._process_parts(list(args)))

    def _process_parts(self, parts: Union[List[Composable], "Composed"]) -> Generator[_Singleton, None, None]:
        for part in iter(parts):
            if isinstance(part, Composed):
                for part in self._process_parts(part):
                    yield part
            elif isinstance(part, _Singleton):
                yield part
            else:
                raise TypeError("Unexpected type " "%s" "", part.__class__.__name__)

    def __iter__(self) -> Generator[_Singleton, None, None]:
        for part in iter(self._parts):
            yield part

    def __len__(self) -> int:
        return len(self._parts)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Composable):
            return False
        return list(iter(self.compose())) == list(iter(other.compose()))

    def __str__(self) -> str:
        return "Composed(" + (" + ".join([str(p) for p in self._parts])) + ")"

    def __repr__(self) -> str:
        return self.__str__()


class SQL(_Singleton):
    """A sql statememt

    Args:
        statement (str): query text
    """

    def __init__(self, statement: str) -> None:
        super().__init__(statement)

    def join(self, iter: Sequence[Composable]) -> Composed:
        """Joins basic query building blocks, such as SQL, Literal, Identifier.

        Args:
            iter: an iterable containing Composable elements

        Returns:
            Composed: a composed query object

        Example:
            sql = SQL("SELECT col FROM ") + SQL(".").join(Identifier("myschema"),Identifier("mytable"))
        """
        parts = []
        for i in range(len(iter)):
            parts.append(iter[i])
            if i < len(iter) - 1:
                parts.append(self)
        return Composed(*parts)

    def format(self, *args: Composable, **kwargs: Composable) -> Composed:
        """Format a query that contains a string formatter argument notation. Format
        values can be one of: Identifier, Literal.

        Args:
            *args: for {} or {0} notation
            *kwargs: for {keyword} notation

        Returns:
            Composed: a composed query object

        Example:
            sql = SQL("SELECT {column} FROM {table} WHERE id = {value}").format(
                column=Identifier("col1"),
                table=Identifier("mytable"),
                value=Literal(1)
            )
        """
        formatter = string.Formatter()
        parts = []
        field_counter = -1
        for text, field_name, format_spec, conversion in formatter.parse(str(self._value)):
            parts.append(SQL(text))
            if field_name is not None:
                if field_name:
                    if field_counter >= 0:
                        raise ValueError(
                            "cannot switch from automatic field numbering to manual field specification"
                        )
                    parts.append(formatter.get_value(field_name, args, kwargs))
                else:
                    field_counter += 1
                    parts.append(formatter.get_value(field_counter, args, kwargs))
        return Composed(*parts)


class Identifier(_Singleton):
    """A database Identifier. Ensures proper quoting for identifiers.

    Args:
        identifier (str): identifier name
    """

    def __init__(self, identifier: str) -> None:
        super().__init__(identifier)


class Literal(_Singleton):
    """A Literal query parameter. Ensures proper quoting for literal values.

    Args:
        literal (object): literal value
    """

    def __init__(self, literal: object) -> None:
        super().__init__(literal)

    def __str__(self) -> str:
        if isinstance(self._value, str):
            return super().__str__()
        return self.__class__.__name__ + f"({self._value})"



"""Errors raised by the adapters"""


class AdapterError(Exception):
    """Generic adapter exception"""

    pass


class ProgrammingError(AdapterError):
    """Programming adapter exception raised by the provider"""

    pass


class NoResultsToFetch(AdapterError):
    """Raised when query did not return any results"""

    pass



"""Postgresql table objects"""
from typing import TYPE_CHECKING, Optional
from ..sql import SQL, Identifier
from ..errors import *
from .. import util
from . import generic


if TYPE_CHECKING:
    from ..cluster import Cluster


class Table(generic._DynamicObject, generic._CollectionChild):
    """Postgres Table object. Represents a table object on a Postgres server.

    Args:
        name (str): table name
        cluster (str): Postgres cluster object
        schema (str): schema name
        owner (str): table owner
        oid (int): table OID
        parent (TableCollection): parent collection

    Attributes:
        name (str): Table name
        cluster (str): Postgres cluster object
        schema (str): Schema name
        owner (str): Table owner
        tablespace (str): Tablespace
        row_security (bool): Whether the row security is enabled
        oid (int): Table OID
    """

    def __init__(
        self,
        name: str,
        schema: str = "public",
        owner: str = None,
        cluster: "Cluster" = None,
        parent: "TableCollection" = None,
        oid: Optional[int] = None,
    ):
        """Initialize a new Table object"""
        super().__init__(kind="TABLE", cluster=cluster, oid=oid, name=name, schema=schema)
        generic._CollectionChild.__init__(self, parent=parent)
        self._schema: str = schema
        self._owner = owner
        self._tablespace: Optional[str] = None
        self._row_security: bool = False

    def drop(self, cascade: bool = False):
        """Drops the table from the Postgres cluster

        Args:
            cascade (bool): drop dependent objects
        """
        sql = SQL("DROP TABLE {table}").format(table=self._sql_fqn())
        if cascade:
            sql += SQL(" CASCADE")
        self.cluster.execute(sql)

    def refresh(self):
        """Re-initializes the object, refreshing its properties from Postgres cluster"""
        super().refresh()
        if not self._ephemeral:
            sql = util.get_sql("get_table") + SQL(" WHERE c.oid = %s")
            result = self.cluster.execute(sql, (self.oid,))
            if not result:
                raise PostgresError("Table with oid %s was not found", self.oid)
            mapper = _TableMapper(result[0])
            mapper.map(self)

    @property
    def tablespace(self):
        return self._tablespace

    @tablespace.setter
    def tablespace(self, value: str):
        if self._tablespace != value:
            self._changes["tablespace"] = generic._SQLChange(
                obj=self,
                sql=SQL("ALTER TABLE {fqn} SET TABLESPACE {tablespace}").format(
                    fqn=self._sql_fqn(), tablespace=Identifier(value)
                ),
            )
            self._tablespace = value

    @property
    def row_security(self):
        return self._row_security

    @row_security.setter
    def row_security(self, value: bool):
        if self._row_security != value:
            keyword = "ENABLE" if value else "DISABLE"
            self._changes["row_security"] = generic._SQLChange(
                obj=self,
                sql=SQL(f"ALTER TABLE {{fqn}} {keyword} ROW LEVEL SECURITY").format(fqn=self._sql_fqn()),
            )
            self._row_security = value

    @property
    def owner(self) -> Optional[str]:
        return self._owner

    @owner.setter
    def owner(self, owner: str):
        generic._set_ephemeral_attr(self, "owner", owner)

    @property
    def schema(self) -> str:
        return self._schema

    @schema.setter
    def schema(self, schema: str):
        generic._set_ephemeral_attr(self, "schema", schema)

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str):
        generic._set_ephemeral_attr(self, "name", name)


class _TableMapper(generic._BaseObjectMapper[Table]):
    """Maps out a resultset from a database query to a table object"""

    attributes = [
        "name",
        "owner",
        "schema",
        "tablespace",
        "row_security",
        "oid",
    ]


class TableCollection(generic._BaseCollection[Table]):
    """An iterable collection of tables indexed by table name.
    For tables outside of the 'public' schema, index becomes "schemaname.tablename".
    """

    def __init__(self, cluster: "Cluster"):
        super().__init__(cluster=cluster)
        if cluster:
            self.refresh()

    def refresh(self):
        """Resets any pending changes and refreshes the list of child objects from the cluster"""
        super().refresh()
        sql = util.get_sql("get_table")
        result = self.cluster.execute(sql)
        for mapper in [_TableMapper(x) for x in result]:
            self[self._index(name=mapper["name"], schema=mapper["schema"])] = mapper.map(
                Table(
                    cluster=self.cluster,
                    name=mapper["name"],
                    schema=mapper["schema"],
                    parent=self,
                    oid=mapper["oid"],
                )
            )



"""Replication slots. Represent replication slots of the Postgres cluster."""
from typing import TYPE_CHECKING, Optional, Union
from ..adapters import AdapterError
from ..sql import SQL, Literal, Composable
from ..errors import PostgresError
from .. import util
from . import generic

if TYPE_CHECKING:
    from ..cluster import Cluster


class ReplicationSlot(generic._DynamicObject, generic._CollectionChild):
    """
    Postgres ReplicationSlot object. Represents a replication slot on a Postgres server.

    Args:
        name (str): name of the replication slot
        cluster (str): Postgres cluster object
        plugin (str): Replication slot plugin name
        parent (ReplicationSlotCollection): parent collection

    Attributes:
        name (str): Replication slot name
        cluster (str): Postgres cluster object
        plugin (str): Replication slot plugin name
        slot_type (str): Replication slot type
        database (str): Database name in which the slot is created
        temporary (bool): Whether the slot is temporary
        is_active (bool): Whether the slot is active
        active_pid (int): Active PID connected to the slot
        xmin (str): Slot xmin
        catalog_xmin (str): Slot catalog_xmin
        restart_lsn (str): Slot restart_lsn
        confirmed_flush_lsn (str): Slot confirmed_flush_lsn
    """

    def __init__(
        self,
        name: str,
        plugin: str,
        cluster: "Cluster" = None,
        parent: "ReplicationSlotCollection" = None,
    ):
        """Initialize a new ReplicationSlot object"""
        super().__init__(cluster=cluster, name=name, kind="REPLICATION SLOT")
        generic._CollectionChild.__init__(self, parent=parent)
        self._plugin = plugin
        self._slot_type = "logical"
        self._database = None
        self._temporary = False
        self._is_active = None
        self._active_pid = None
        self._xmin = None
        self._catalog_xmin = None
        self._restart_lsn = None
        self._confirmed_flush_lsn = None

    # properties
    @property
    def plugin(self) -> Optional[str]:
        return self._plugin

    @property
    def slot_type(self) -> str:
        return self._slot_type

    @property
    def database(self) -> Optional[str]:
        return self._database

    @property
    def temporary(self) -> bool:
        return self._temporary

    @property
    def is_active(self) -> Optional[bool]:
        return self._is_active

    @property
    def active_pid(self) -> Optional[int]:
        return self._active_pid

    @property
    def xmin(self) -> Optional[int]:
        return self._xmin

    @property
    def catalog_xmin(self) -> Optional[int]:
        return self._catalog_xmin

    @property
    def restart_lsn(self) -> Optional[int]:
        return self._restart_lsn

    @property
    def confirmed_flush_lsn(self) -> Optional[int]:
        return self._confirmed_flush_lsn

    # methods
    def drop(self, retries: int = 10) -> None:
        """Drops the replication slot from the Postres cluster

        Args:
            retries (int): number of retries after a failure. Useful when a subscriber is programmed to reconnect.
        """
        attempts = 0
        while True:
            self.refresh()
            if self.active_pid:
                self.disconnect()
            try:
                self.cluster.execute(SQL("SELECT pg_drop_replication_slot(%s)"), self.name)
                break
            except AdapterError:
                if attempts >= retries:
                    raise
                attempts += 1

    def create(self) -> None:
        """Creates a replication slot in the currently connected database based on currently configured attributes."""
        self.cluster.execute(self.script(as_composable=True))
        self.refresh()

    def script(self, as_composable: bool = False) -> Union[str, Composable]:
        """Generate a database creation script.

        Args:
            as_composable (bool): return Composable object instead of plain text

        Returns:
            Union[str, Composable]: replication slot creation script
        """
        sql = SQL("SELECT pg_create_logical_replication_slot({}, {})")

        command = sql.format(Literal(self.name), Literal(self.plugin))
        if as_composable:
            return command
        else:
            with self.cluster.adapter.cursor() as cur:
                return cur.mogrify(command)

    def disconnect(self):
        """Terminates the active pid of the replication slot"""
        self.cluster.execute(
            SQL(
                "SELECT pg_terminate_backend(active_pid) FROM pg_catalog.pg_replication_slots WHERE slot_name = %s"
            ),
            self.name,
        )

    def refresh(self):
        """Re-initializes the object, refreshing its properties"""
        super().refresh()
        sql = util.get_sql("get_replication_slot") + SQL(" WHERE slot_name = %s")
        result = self.cluster.execute(sql, self._name)
        if not result:
            raise PostgresError("Replication slot not found")
        mapper = _ReplicationSlotMapper(result[0])
        mapper.map(self)


class _ReplicationSlotMapper(generic._BaseObjectMapper[ReplicationSlot]):
    """Maps out a resultset from a replication slot query to a replication slot object"""

    attributes = [
        "name",
        "plugin",
        "slot_type",
        "database",
        "temporary",
        "is_active",
        "active_pid",
        "xmin",
        "catalog_xmin",
        "restart_lsn",
        "confirmed_flush_lsn",
    ]


class ReplicationSlotCollection(generic._BaseCollection[ReplicationSlot]):
    """An iterable collection of replication slots indexed by slot name."""

    def __init__(self, cluster: "Cluster"):
        super().__init__(cluster=cluster)
        if cluster:
            self.refresh()

    def refresh(self):
        """Resets any pending changes and retrieves objects from the cluster"""

        super().refresh()
        sql = util.get_sql("get_replication_slot") + SQL(" ORDER BY slot_name")
        result = self.cluster.execute(sql)
        for mapper in [_ReplicationSlotMapper(x) for x in result]:
            self[mapper["name"]] = mapper.map(
                ReplicationSlot(
                    cluster=self.cluster,
                    name=mapper["name"],
                    plugin=mapper["plugin"],
                    parent=self,
                )
            )

    def new(
        self,
        name: str,
        plugin: str,
    ) -> ReplicationSlot:
        """Create a replication on on the current Postgres cluster. The object
        is created ephemeral and either needs to be added to the slot collection,
        or .create() needs to be executed.

        Args:
            name (str): name of the replication slot
            cluster (str): Postgres cluster object
            plugin (str): Replication slot plugin name
            temporary (bool): Whether the slot is temporary
        """
        return ReplicationSlot(
            cluster=self.cluster,
            name=name,
            plugin=plugin,
            parent=self,
        )

    def add(self, slot: ReplicationSlot):
        """Adds the replication slot object to the collection, simultaneously creating it on the cluster

        Args:
            slot (ReplicationSlot): initialized replication slot object
        """
        slot.cluster = self.cluster
        slot._parent = self
        slot.create()
        self[slot.name] = slot



"""Postgresql roles"""
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, Optional, Union
from ..adapters import AdapterError
from ..sql import SQL, Composable, Literal, Identifier
from ..errors import *
from .. import util
from . import generic


if TYPE_CHECKING:
    from ..cluster import Cluster


class Role(generic._DynamicObject, generic._CollectionChild):
    """
    Postgres Role object. Represents a role object on a Postgres server.

    Args:
        name (str): name of the role
        password (str): role password with which it will be created
        cluster (str): Postgres cluster object
        superuser (bool): SUPERUSER permissions
        inherit (bool): INHERIT permissions
        createrole (bool): CREATEROLE permissions
        createdb (bool): CREATEDB permissions
        login (bool): LOGIN permissions
        replication (bool): REPLICATION permissions
        bypassrls (bool): BYPASSRLS enabled
        connection_limit (int): Role connection limit
        valid_until (datetime): Expires on this date
        oid (int): Role OID
        parent (DatabaseCollection): parent collection

    Attributes:
        name (str): Table name
        cluster (str): Postgres cluster object
        superuser (bool): SUPERUSER permissions
        inherit (bool): INHERIT permissions
        createrole (bool): CREATEROLE permissions
        createdb (bool): CREATEDB permissions
        login (bool): LOGIN permissions
        replication (bool): REPLICATION permissions
        bypassrls (bool): BYPASSRLS enabled
        connection_limit (int): Role connection limit
        valid_until (datetime): Expires on this date
        oid (int): Role OID
    """

    def __init__(
        self,
        name: str,
        password: str = None,
        cluster: "Cluster" = None,
        superuser: bool = False,
        inherit: bool = True,
        createrole: bool = False,
        createdb: bool = False,
        login: bool = False,
        replication: bool = False,
        bypassrls: bool = False,
        connection_limit: int = -1,
        valid_until: datetime = None,
        oid: int = None,
        parent: "RoleCollection" = None,
    ):
        super().__init__(kind="ROLE", cluster=cluster, oid=oid, name=name)
        generic._CollectionChild.__init__(self, parent=parent)
        self._password = password
        self._cluster = cluster
        self._superuser = superuser
        self._inherit = inherit
        self._createrole = createrole
        self._createdb = createdb
        self._login = login
        self._replication = replication
        self._bypassrls = bypassrls
        self._connection_limit = connection_limit
        self._valid_until = valid_until
        self._oid = oid

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name: str):
        generic._set_ephemeral_attr(self, "name", name)

    @property
    def superuser(self) -> bool:
        return self._superuser

    @superuser.setter
    def superuser(self, value: bool):
        self._set_permission("superuser", value)

    @property
    def inherit(self) -> bool:
        return self._inherit

    @inherit.setter
    def inherit(self, value: bool):
        self._set_permission("inherit", value)

    @property
    def createrole(self) -> bool:
        return self._createrole

    @createrole.setter
    def createrole(self, value: bool):
        self._set_permission("createrole", value)

    @property
    def createdb(self) -> bool:
        return self._createdb

    @createdb.setter
    def createdb(self, value: bool):
        self._set_permission("createdb", value)

    @property
    def login(self) -> bool:
        return self._login

    @login.setter
    def login(self, value: bool):
        self._set_permission("login", value)

    @property
    def replication(self) -> bool:
        return self._replication

    @replication.setter
    def replication(self, value: bool):
        self._set_permission("replication", value)

    @property
    def bypassrls(self) -> bool:
        return self._bypassrls

    @bypassrls.setter
    def bypassrls(self, value: bool):
        self._set_permission("bypassrls", value)

    @property
    def valid_until(self) -> Optional[datetime]:
        return self._valid_until

    @valid_until.setter
    def valid_until(self, value: Optional[datetime]):
        self._set_attribute("valid_until", str(value))

    @property
    def connection_limit(self) -> int:
        return self._connection_limit

    @connection_limit.setter
    def connection_limit(self, value: Optional[int]):
        self._set_attribute("connection_limit", value)

    def _set_attribute(self, attr: str, value: Any):
        if getattr(self, f"_{attr}") != value:
            sql = SQL(f"ALTER ROLE {{fqn}} {attr.upper().replace('_', ' ')} {'%s' if value else 'DEFAULT'}")
            params = (value,) if value else None
            stmt = sql.format(fqn=self._sql_fqn())
            self._changes[attr] = generic._SQLChange(obj=self, sql=stmt, params=params)
            setattr(self, f"_{attr}", value)

    def _set_permission(self, attr: str, value: bool):
        if getattr(self, f"_{attr}") != value:
            permission = ("" if value else "NO") + attr.upper()
            stmt = SQL(f"ALTER ROLE {{fqn}} {permission}").format(fqn=self._sql_fqn())
            self._changes[attr] = generic._SQLChange(obj=self, sql=stmt)
            setattr(self, f"_{attr}", permission)

    def script(self, as_composable: bool = False) -> Union[str, Composable]:
        """Scripts out a role.

        Args:
            as_composable (bool): return Composable object instead of plain text

        Returns:
            Union[str, Composable]: role creation script
        """
        permission_list = {
            "SUPERUSER": self.superuser,
            "CREATEDB": self.createdb,
            "CREATEROLE": self.createrole,
            "INHERIT": self.inherit,
            "LOGIN": self.login,
            "REPLICATION": self.replication,
            "BYPASSRLS": self.bypassrls,
        }
        sql = "CREATE ROLE {role}"
        params: Dict[str, Composable] = {"role": self._sql_fqn()}
        password = self._password if self._password else self.get_password_md5()
        if password:
            sql += " PASSWORD {password}"
            params["password"] = Literal(password)
        for permission in permission_list.keys():
            sql += " " + (permission if permission_list[permission] else f"NO{permission}")
        if self.connection_limit is not None:
            sql += " CONNECTION LIMIT {limit}"
            params["limit"] = Literal(self.connection_limit)
        if self.valid_until is not None:
            sql += " VALID UNTIL {valid_until}"
            params["valid_until"] = Literal(str(self.valid_until))

        command = SQL(sql).format(**params)
        if as_composable:
            return command
        else:
            with self.cluster.adapter.cursor() as cur:
                return cur.mogrify(command)

    def create(self) -> None:
        """Create a role on the Postgres cluster"""
        self.cluster.execute(self.script(as_composable=True))
        sql = util.get_sql("get_role") + SQL(" WHERE rolname = %s")
        _RoleMapper(self.cluster.execute(sql, self.name)[0]).map(self)

    def drop(self, force: bool = False) -> None:
        """Drops the role connected to this object

        Args:
            force (bool): disallow new connections and terminate all existing ones for the role before executing the drop statement
        """
        if force:
            try:
                self.cluster.execute(SQL("ALTER ROLE {fqn} NOLOGIN").format(fqn=self._sql_fqn()))
                self.cluster.terminate(roles=[self.name])
            except AdapterError:
                if self.login:
                    self.cluster.execute(SQL("ALTER ROLE {fqn} LOGIN").format(fqn=self._sql_fqn()))
                raise
        self.cluster.execute(SQL("DROP ROLE {fqn}").format(fqn=self._sql_fqn()))

    def get_password_md5(self) -> Optional[str]:
        """Returns md5 password hash for the role"""
        sql = SQL("SELECT rolpassword FROM pg_catalog.pg_authid WHERE rolname = %s")
        result = self.cluster.execute(sql, self.name)
        if not result:
            raise PostgresError("Role %s is not found", self.name)
        return result[0][0]

    def refresh(self):
        """Re-initializes the object, refreshing its properties"""
        super().refresh()
        if not self._ephemeral:
            sql = util.get_sql("get_role") + SQL(" WHERE oid = %s")
            result = self.cluster.execute(sql, self._oid)
            if not result:
                raise PostgresError("Role not found")
            mapper = _RoleMapper(result[0])
            mapper.map(self)

    def change_password(self, password: str) -> None:
        """Changes the role's password

        Args:
            password (str): new password
        """
        sql = SQL("ALTER ROLE {role} PASSWORD {password}").format(
            role=Identifier(self.name), password=Literal(str(password))
        )
        self.cluster.execute(sql)


class _RoleMapper(generic._BaseObjectMapper[Role]):
    """Maps out a resultset from a database query to a role object"""

    attributes = [
        "name",
        "superuser",
        "inherit",
        "createrole",
        "createdb",
        "login",
        "replication",
        "connection_limit",
        "valid_until",
        "bypassrls",
        "oid",
    ]


class RoleCollection(generic._BaseCollection[Role]):
    """An iterable collection of roles indexed by role name.

    Args:
        cluster (str): Postgres cluster object

    Attributes:
        cluster (str): Postgres cluster object
    """

    def __init__(self, cluster):
        super().__init__(cluster=cluster)
        if cluster:
            self.refresh()

    def refresh(self):
        """Resets any pending changes and retrieves objects from the cluster"""
        super().refresh()
        sql = util.get_sql("get_role") + SQL(" ORDER BY rolname")
        result = self.cluster.execute(sql)
        for mapper in [_RoleMapper(x) for x in result]:
            self[mapper["name"]] = mapper.map(Role(cluster=self.cluster, name=mapper["name"], parent=self))

    def new(
        self,
        name: str,
        password: str = None,
        superuser: bool = False,
        inherit: bool = True,
        createrole: bool = False,
        createdb: bool = False,
        login: bool = False,
        replication: bool = False,
        bypassrls: bool = False,
        connection_limit: int = -1,
        valid_until: datetime = None,
    ) -> Role:
        """Create a role object on the current Postgres cluster. The object
        is created ephemeral and either needs to be added to the role collection,
        or .create() needs to be executed.

        Args:
            name (str): name of the role
            password (str): role password with which it will be created
            superuser (bool): SUPERUSER permissions
            inherit (bool): INHERIT permissions
            createrole (bool): CREATEROLE permissions
            createdb (bool): CREATEDB permissions
            login (bool): LOGIN permissions
            replication (bool): REPLICATION permissions
            bypassrls (bool): BYPASSRLS enabled
            connection_limit (int): Role connection limit
            valid_until (datetime): Expires on this date
        """
        return Role(
            cluster=self.cluster,
            name=name,
            parent=self,
            password=password,
            superuser=superuser,
            inherit=inherit,
            createrole=createrole,
            createdb=createdb,
            login=login,
            replication=replication,
            bypassrls=bypassrls,
            connection_limit=connection_limit,
            valid_until=valid_until,
        )

    def add(self, role: Role):
        """Adds the role object to the collection, simultaneously creating it on the cluster

        Args:
            role (Role): initialized role object
        """
        role.cluster = self.cluster
        role._parent = self
        role.create()
        self[role.name] = role



"""Postgresql procedure objects"""
from typing import TYPE_CHECKING, Dict, List, Optional, Type
from ..sql import SQL, Identifier, Composable
from ..errors import *
from .. import util
from . import generic


if TYPE_CHECKING:
    from ..cluster import Cluster


class ProcedureKind(generic.AliasEnum):
    """Postgres procedure kinds"""

    FUNCTION = "FUNCTION"
    PROCEDURE = "PROCEDURE"
    AGGREGATE = "AGGREGATE"
    WINDOW_FUNCTION = "WINDOW FUNCTION"


class Volatility(generic.AliasEnum):
    """Postgres function volatility, as described in
    https://www.postgresql.org/docs/current/xfunc-volatility.html
    """

    IMMUTABLE = "i"
    STABLE = "s"
    VOLATILE = "v"


class ParallelSafety(generic.AliasEnum):
    """Postgres parallel safety, as described in
    https://www.postgresql.org/docs/current/parallel-safety.html
    """

    SAFE = "s"
    RESTRICTED = "r"
    UNSAFE = "u"


class _BaseProcedure(generic._DynamicObject, generic._CollectionChild):
    """Postgres Procedure base object. Represents a stored procedure,
    function, window function or aggregate on a Postgres server.

    Args:
        name (str): name of the procedure
        cluster (str): Postgres cluster object
        schema (str): Schema name
        owner (str): Procedure owner
        kind (str): Procedure kind (FUNCTION/PROCEDURE/AGGREGATE/WINDOW FUNCTION)
        security_definer (bool): Whether the function is a security definer
        leak_proof (bool): Whether the function is leak proof
        strict (bool): Whether the function is strict
        volatility (Volatility): Volatility mode (IMMUTABLE/STABLE/VOLATILE)
        parallel_mode (ParallelSafety): Parallel mode (SAFE/RESTRICTED/UNSAFE)
        argument_types (List[str]): A list of argument types, if any
        oid (int): Procedure OID
        parent (ProcedureCollection): parent collection

    Attributes:
        name (str): Procedure name
        cluster (str): Postgres cluster object
        schema (str): Schema name
        owner (str): Procedure owner
        kind (str): Procedure kind (FUNCTION/PROCEDURE/AGGREGATE/WINDOW FUNCTION)
        security_definer (bool): Whether the function is a security definer
        leak_proof (bool): Whether the function is leak proof
        strict (bool): Whether the function is strict
        volatility (Volatility): Volatility mode (IMMUTABLE/STABLE/VOLATILE)
        parallel_mode (ParallelSafety): Parallel mode (SAFE/RESTRICTED/UNSAFE)
        argument_types (List[str]): A list of argument types, if any
        oid (int): Procedure OID
    """

    def __init__(
        self,
        name: str,
        kind: str,
        schema: str = "public",
        argument_types: List[str] = None,
        oid: Optional[int] = None,
        parent: "ProcedureCollection" = None,
        owner: str = None,
        cluster: "Cluster" = None,
        language: str = "sql",
        security_definer: bool = False,
        strict: bool = False,
        leak_proof: bool = False,
        volatility: Volatility = Volatility.VOLATILE,
        parallel_mode: ParallelSafety = ParallelSafety.UNSAFE,
    ):
        super().__init__(kind=kind, cluster=cluster, oid=oid, name=name, schema=schema)
        generic._CollectionChild.__init__(self, parent=parent)
        self._language = language
        self._owner = owner
        self._security_definer = security_definer
        self._leak_proof = leak_proof
        self._strict = strict
        self._volatility = Volatility(volatility)
        self._parallel_mode = ParallelSafety(parallel_mode)
        self._argument_types = argument_types

    def _sql_fqn(self) -> Composable:
        """In a form of: schema.function (arg1, arg2)"""
        fqn = super()._sql_fqn()
        if self._argument_types:
            fqn = fqn + SQL(" (") + SQL(",").join([Identifier(x) for x in self._argument_types]) + SQL(")")
        return fqn

    @property
    def security_definer(self) -> bool:
        return self._security_definer

    @property
    def leak_proof(self) -> bool:
        return self._leak_proof

    @property
    def strict(self) -> bool:
        return self._strict

    @property
    def volatility(self) -> Volatility:
        return self._volatility

    @property
    def parallel_mode(self) -> ParallelSafety:
        return self._parallel_mode

    @property
    def argument_types(self) -> Optional[List[str]]:
        return self._argument_types

    @property
    def language(self) -> str:
        return self._language

    def drop(self, cascade: bool = False):
        """Drops the procedure from the database

        Args:
            cascade (bool): cascade dependent objects
        """

        sql = SQL(f"DROP {self.kind} {{procedure}}").format(procedure=self._sql_fqn())
        if cascade:
            sql += SQL(" CASCADE")
        self.cluster.execute(sql)

    def refresh(self):
        """Re-initialize the object, refreshing its properties from the database"""
        super().refresh()
        if not self._ephemeral:
            sql = util.get_sql("get_procedure", self.cluster.version) + SQL(" AND p.oid = %s")
            result = self.cluster.execute(sql, (self.oid))
            if not result:
                raise PostgresError("Procedure with oid %s was not found", self.oid)
            mapper = _ProcedureMapper(result[0])
            mapper.map(self)

    @property
    def owner(self) -> Optional[str]:
        return self._owner

    @owner.setter
    def owner(self, owner: str):
        generic._set_ephemeral_attr(self, "owner", owner)

    @property
    def schema(self) -> Optional[str]:
        return self._schema

    @schema.setter
    def schema(self, schema: str):
        generic._set_ephemeral_attr(self, "schema", schema)

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str):
        generic._set_ephemeral_attr(self, "name", name)


class Procedure(_BaseProcedure):
    """Postgres Procedure object. Represents a stored procedure on a Postgres server.

    Args:
        name (str): name of the procedure
        cluster (str): Postgres cluster object
        schema (str): Schema name
        owner (str): Procedure owner
        security_definer (bool): Whether the procedure is a security definer
        leak_proof (bool): Whether the procedure is leak proof
        strict (bool): Whether the procedure is strict
        volatility (Volatility): Volatility mode (IMMUTABLE/STABLE/VOLATILE)
        parallel_mode (ParallelSafety): Parallel mode (SAFE/RESTRICTED/UNSAFE)
        argument_types (List[str]): A list of argument types, if any
        oid (int): Procedure OID

    Attributes:
        name (str): Procedure name
        cluster (str): Postgres cluster object
        schema (str): Schema name
        owner (str): Procedure owner
        kind (str): Procedure kind (FUNCTION/PROCEDURE/AGGREGATE/WINDOW FUNCTION)
        security_definer (bool): Whether the function is a security definer
        leak_proof (bool): Whether the function is leak proof
        strict (bool): Whether the function is strict
        volatility (Volatility): Volatility mode (IMMUTABLE/STABLE/VOLATILE)
        parallel_mode (ParallelSafety): Parallel mode (SAFE/RESTRICTED/UNSAFE)
        argument_types (List[str]): A list of argument types, if any
        oid (int): Procedure OID
    """

    def __init__(self, *args, **kwargs):
        super().__init__(kind=ProcedureKind.PROCEDURE.name, *args, **kwargs)


class Function(_BaseProcedure):
    """Postgres Function object. Represents a function on a Postgres server.

    Args:
        name (str): name of the function
        cluster (str): Postgres cluster object
        schema (str): Schema name
        owner (str): Function owner
        security_definer (bool): Whether the function is a security definer
        leak_proof (bool): Whether the function is leak proof
        strict (bool): Whether the function is strict
        volatility (Volatility): Volatility mode (IMMUTABLE/STABLE/VOLATILE)
        parallel_mode (ParallelSafety): Parallel mode (SAFE/RESTRICTED/UNSAFE)
        argument_types (List[str]): A list of argument types, if any
        oid (int): Procedure OID

    Attributes:
        name (str): Procedure name
        cluster (str): Postgres cluster object
        schema (str): Schema name
        owner (str): Procedure owner
        security_definer (bool): Whether the function is a security definer
        leak_proof (bool): Whether the function is leak proof
        strict (bool): Whether the function is strict
        volatility (Volatility): Volatility mode (IMMUTABLE/STABLE/VOLATILE)
        parallel_mode (ParallelSafety): Parallel mode (SAFE/RESTRICTED/UNSAFE)
        argument_types (List[str]): A list of argument types, if any
        oid (int): Procedure OID
    """

    def __init__(self, *args, **kwargs):
        super().__init__(kind=ProcedureKind.FUNCTION.name, *args, **kwargs)


class WindowFunction(_BaseProcedure):
    """Postgres Window Function object. Represents a window function on a Postgres server.

    Args:
        name (str): name of the function
        cluster (str): Postgres cluster object
        schema (str): Schema name
        owner (str): Function owner
        security_definer (bool): Whether the function is a security definer
        leak_proof (bool): Whether the function is leak proof
        strict (bool): Whether the function is strict
        volatility (Volatility): Volatility mode (IMMUTABLE/STABLE/VOLATILE)
        parallel_mode (ParallelSafety): Parallel mode (SAFE/RESTRICTED/UNSAFE)
        argument_types (List[str]): A list of argument types, if any
        oid (int): Procedure OID

    Attributes:
        name (str): Procedure name
        cluster (str): Postgres cluster object
        schema (str): Schema name
        owner (str): Procedure owner
        security_definer (bool): Whether the function is a security definer
        leak_proof (bool): Whether the function is leak proof
        strict (bool): Whether the function is strict
        volatility (Volatility): Volatility mode (IMMUTABLE/STABLE/VOLATILE)
        parallel_mode (ParallelSafety): Parallel mode (SAFE/RESTRICTED/UNSAFE)
        argument_types (List[str]): A list of argument types, if any
        oid (int): Procedure OID
    """

    def __init__(self, *args, **kwargs):
        super().__init__(kind=ProcedureKind.WINDOW_FUNCTION.name, *args, **kwargs)


class Aggregate(_BaseProcedure):
    """Postgres Aggregate object. Represents an aggregate on a Postgres server.

    Args:
        name (str): name of the aggregate function
        cluster (str): Postgres cluster object
        schema (str): Schema name
        owner (str): Function owner
        security_definer (bool): Whether the function is a security definer
        leak_proof (bool): Whether the function is leak proof
        strict (bool): Whether the function is strict
        volatility (Volatility): Volatility mode (IMMUTABLE/STABLE/VOLATILE)
        parallel_mode (ParallelSafety): Parallel mode (SAFE/RESTRICTED/UNSAFE)
        argument_types (List[str]): A list of argument types, if any
        oid (int): Procedure OID


    Attributes:
        name (str): Procedure name
        cluster (str): Postgres cluster object
        schema (str): Schema name
        owner (str): Procedure owner
        security_definer (bool): Whether the function is a security definer
        leak_proof (bool): Whether the function is leak proof
        strict (bool): Whether the function is strict
        volatility (Volatility): Volatility mode (IMMUTABLE/STABLE/VOLATILE)
        parallel_mode (ParallelSafety): Parallel mode (SAFE/RESTRICTED/UNSAFE)
        argument_types (List[str]): A list of argument types, if any
        oid (int): Procedure OID
    """

    def __init__(self, *args, **kwargs):
        super().__init__(kind=ProcedureKind.AGGREGATE.name, *args, **kwargs)


class _ProcedureMapper(generic._BaseObjectMapper[_BaseProcedure]):
    """Maps out a resultset from a database query to a procedure object"""

    attributes = [
        "oid",
        "name",
        "schema",
        "owner",
        "language",
        "kind",
        "security_definer",
        "leak_proof",
        "strict",
        "volatility",
        "parallel_mode",
        "argument_types",
    ]
    exclude = ["kind", "volatility", "parallel_mode"]

    def map(self, obj: _BaseProcedure) -> _BaseProcedure:
        """Assigns attributes to the Postgres procedure based on
        the definition resultset and returns the object.

        Args:
            obj (_ProcedureBase): Postgres procedure object
        """
        super().map(obj)
        obj._volatility = Volatility(self["volatility"])
        obj._parallel_mode = ParallelSafety(self["parallel_mode"])

        return obj


class ProcedureVariations(List[_BaseProcedure], generic._ClusterBound):
    """A list of procedures with the same name, but variable argument sets."""

    def __init__(self, cluster: "Cluster"):
        super().__init__()
        generic._ClusterBound.__init__(self, cluster=cluster)


_procedure_kinds: Dict[str, Type[_BaseProcedure]] = {
    "f": Function,
    "p": Procedure,
    "w": WindowFunction,
    "a": Aggregate,
}


class ProcedureCollection(generic._BaseCollection[ProcedureVariations]):
    """An iterable collection of procedure variations indexed by procedure name.
    For procedures outside of the 'public' schema, index becomes "schemaname.procedurename".
    Procedure variations contains all procedures with the same name, yet different
    set of arguments.
    """

    def __init__(self, cluster: "Cluster"):
        super().__init__(cluster=cluster)
        if cluster:
            self.refresh()

    def refresh(self):
        """Resets any pending changes and retrieves objects from the cluster"""
        super().refresh()
        sql = util.get_sql("get_procedure", self.cluster.version) + SQL(
            " ORDER BY proname, schemaname, proargtypes"
        )
        result = self.cluster.execute(sql)
        if result:
            mappers = [_ProcedureMapper(x) for x in result]
            # group by schema and name for index access, then put each group into a ProcedureVariations list
            grouped_by_index = util.group_by(
                lambda x: self._index(name=x["name"], schema=x["schema"]), mappers
            )
            for key, mapper_group in grouped_by_index.items():
                self[key] = ProcedureVariations(cluster=self.cluster)
                for mapper in mapper_group:
                    kind = mapper["kind"]
                    try:
                        proc_class = _procedure_kinds[kind]
                    except KeyError:
                        raise PostgresError(f"Unknown procedure kind: {kind}")
                    proc = proc_class(
                        name=mapper["name"],
                        schema=mapper["schema"],
                        oid=mapper["oid"],
                        cluster=self.cluster,
                        parent=self,
                    )
                    self[key].append(mapper.map(proc))



"""Internal generic objects shared between other objects."""

from abc import abstractmethod
from enum import Enum
import re
from typing import Any, Dict, Generic, Iterator, List, Optional, TYPE_CHECKING, Tuple, TypeVar
from pgmob import errors
from pgmob._decorators import deprecated
from pgmob.sql import SQL, Identifier, Composable

if TYPE_CHECKING:
    from ..cluster import Cluster


T = TypeVar("T")


class _BasePostgresObject(object):
    """Base class for any Postgres object with oid.

    Args:
        oid (Optional[int]): object id

    Attributes:
        oid (Optional[int]): object id
    """

    def __init__(self, oid: Optional[int] = None):
        self._oid = oid

    # properties
    @property
    def oid(self) -> Optional[int]:
        return self._oid

    @oid.setter
    def oid(self, value: int):
        if not self._ephemeral:
            raise errors.PostgresError("The object is not ephemeral. Can't set oid.")
        self._oid = value

    @property
    def _ephemeral(self) -> bool:
        return self._oid is None or self._oid <= 0


class _BaseObjectMapper(Generic[T]):
    """Maps the resultset to a Dynamic Object"""

    attributes: List[str] = []
    exclude: List[str] = []

    def __init__(self, definition: tuple):
        self.definition = definition

    def __getitem__(self, key: str) -> object:
        return self.definition[self.attributes.index(key)]

    def map(self, obj: T) -> T:
        """Assigns attributes to the Postgres object based on the definition resultset
        and returns the object.

        Args:
            obj (_BasePostgresObject): Postgres object
        """
        for i in range(len(self.attributes)):
            if self.attributes[i] not in self.exclude:
                setattr(obj, f"_{self.attributes[i]}", self.definition[i])

        return obj


class _ClusterBound(object):
    """Object that is attached to a Cluster. Implements cluster retrieval internal function"""

    def __init__(self, cluster: "Cluster" = None):
        self._cluster = cluster

    @property
    def cluster(self) -> "Cluster":
        """Retrieves the Cluster instance bound to the object

        Returns:
            Cluster: Postgres cluster object
        """
        if not self._cluster:
            raise errors.PostgresError("Object is not associated with a cluster")
        return self._cluster

    @cluster.setter
    def cluster(self, value: "Cluster"):
        """Bounds object to a cluster

        Args:
            value (Cluster): Postgres cluster object
        """
        from ..cluster import Cluster

        if value is not None and not isinstance(value, Cluster):
            raise ValueError("%s is not a cluster object", value.__class__)
        self._cluster = value


class Fqn(object):
    """Fully qualified object name. Defined by schema name (optional) and object name.

    Args:
        name (str): object name
        schema (Optional[str]): schema name
    """

    def __init__(self, name: str, schema: Optional[str] = None):
        self.name = name
        self.schema = schema

    def get_identifier(self) -> Composable:
        """Returns a prepared identifier or set of identifiers to insert into the query

        Returns:
            Composable: object identifier
        """
        result: Composable = Identifier(self.name)
        if self.schema:
            result = SQL(".").join([Identifier(self.schema), result])
        return result


class _FqnObject(_BasePostgresObject, _ClusterBound):
    """Base Postgres object that has FQN, bound to a cluster and supports ephemeral changes"""

    def __init__(
        self,
        kind: str,
        name: str,
        oid: Optional[int] = None,
        schema: Optional[str] = None,
        cluster: "Cluster" = None,
    ):
        _BasePostgresObject.__init__(self, oid=oid)
        _ClusterBound.__init__(self, cluster=cluster)
        self._fqn = Fqn(name=name, schema=schema)
        self._name = name
        self._schema = schema
        self._kind = kind

    @property
    def name(self):
        return self._name

    @property
    def schema(self):
        return self._schema

    @property
    def kind(self) -> str:
        return self._kind

    def __repr__(self):
        name = (f"{self.schema}." if self.schema and self.schema != "public" else "") + self.name
        return f"{self.__class__.__name__}('{name}')"

    def _sql_fqn(self) -> Composable:
        return self._fqn.get_identifier()


class _DynamicObject(_FqnObject):
    """Object that supports ephemeral changes. Should be inherited by objects
    that support property change with an eventual .alter() call.
    """

    def __init__(
        self,
        kind: str,
        name: str,
        oid: Optional[int] = None,
        schema: Optional[str] = None,
        cluster: "Cluster" = None,
    ):
        super().__init__(kind=kind, name=name, schema=schema, cluster=cluster, oid=oid)
        self._changes = _ChangeCollection()

    def alter(self):
        """Alters remote object by applying local changes that were queued up for the object"""
        keys = list(self._changes.keys())
        # ensure that alter name is queued up last
        for key in [k for k in keys if k != "name"] + (["name"] if "name" in keys else []):
            self._changes[key].apply()
            self._changes.pop(key)
        self.refresh()

    def refresh(self):
        """Re-initializes the object, refreshing its properties"""
        self._changes = _ChangeCollection()


def _set_ephemeral_attr(obj: _DynamicObject, attr: str, value: Any):
    params = dict(
        fqn=obj._sql_fqn(),
        value=Identifier(value),
    )
    stmt_map = dict(
        owner=SQL(f"ALTER {obj._kind} {{fqn}} OWNER TO {{value}}").format(**params),
        name=SQL(f"ALTER {obj._kind} {{fqn}} RENAME TO {{value}}").format(**params),
        schema=SQL(f"ALTER {obj.kind} {{fqn}} SET SCHEMA {{value}}").format(**params),
        tablespace=SQL(f"ALTER {obj.kind} {{fqn}} SET TABLESPACE {{value}}").format(**params),
    )

    if getattr(obj, f"_{attr}") != value:
        obj._changes[attr] = _SQLChange(obj=obj, sql=stmt_map[attr])
        setattr(obj, f"_{attr}", value)


class MappedCollection(Dict[str, T]):
    """Class implements an iterable dictionary.
    Items are accessed via a key, but when iterated over, acts as a list."""

    def __iter__(self) -> Iterator[T]:  # type: ignore[override]
        for key in self.keys():
            yield self[key]

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.keys()})"


class SortedMappedCollection(Dict[str, T]):
    """Class implements an iterable sorted dictionary.
    Items are accessed via a key, but when iterated over, acts as a sorted list."""

    def __iter__(self) -> Iterator[T]:  # type: ignore[override]
        for key in sorted(self.keys()):
            yield self[key]

    def __repr__(self) -> str:
        return f"{type(self).__name__}({sorted(self.keys())})"


class _BaseCollection(_ClusterBound, SortedMappedCollection[T]):
    """Generic Postgres collection object bound to a cluster."""

    def __init__(self, cluster: "Cluster"):
        SortedMappedCollection.__init__(self)  # type: ignore
        _ClusterBound.__init__(self, cluster=cluster)

    def __setitem__(self, key: str, value: T):
        if not isinstance(value, _ClusterBound):
            raise AttributeError("Unsupported type %s", type(value))
        value.cluster = self.cluster
        SortedMappedCollection.__setitem__(self, key, value)  # type: ignore

    @staticmethod
    def _index(name: str, schema: str):
        """Returns index provided the fqn names"""
        return name if schema == "public" else f"{schema}.{name}"

    def refresh(self):
        """Implements collection refresh, resetting all pending changes and refreshing members.
        Should have an implementation in child classes to load members."""
        self.clear()

    # TODO: implement methods in each child class
    # @abstractmethod
    # def add(self, item: T):
    #     """Abstract method that should be overloaded by each inheriting class that wants to implement add"""
    #     raise NotImplementedError()

    # @abstractmethod
    # def new(self, item: T):
    #     """Abstract method that should be overloaded by each inheriting class that wants to implement new"""
    #     raise NotImplementedError()


class _CollectionChild(object):
    """Defines an object belonging to a Postgres collection"""

    def __init__(self, parent: _BaseCollection = None):
        self._parent = parent

    @property
    def parent(self) -> Optional[_BaseCollection]:
        return self._parent


class _ObjectChange(object):
    def __init__(self, obj: _DynamicObject, task, *args, **kwargs):
        if not isinstance(obj, _DynamicObject):
            raise AttributeError("This object class is not supported")
        self.task = task
        self.object = obj
        self.args = args
        self.kwargs = kwargs

    def apply(self):
        """Applies the change"""
        self.task(*self.args, **self.kwargs)


class _SQLChange(_ObjectChange):
    def __init__(self, obj: _DynamicObject, sql: Composable, params: tuple = None):
        def task():
            obj.cluster.execute(sql, params)

        _ObjectChange.__init__(self, task=task, obj=obj)
        self.sql = sql
        self.params = params


class _ChangeCollection(MappedCollection[_ObjectChange]):
    """An iterable collection of changes indexed by attribute name."""

    def __setitem__(self, key, value):
        if not isinstance(value, _ObjectChange):
            raise AttributeError(f"{value.__class__.__name__} is not of an _ObjectChange type")
        MappedCollection.__setitem__(self, key, value)


class AliasEnum(Enum):
    def __repr__(self):
        return f"{self.__class__.__name__}.{self.name}"

    def __str__(self):
        return self.name



"""Sequence objects"""
from typing import TYPE_CHECKING, Optional
from ..sql import SQL, Literal
from ..errors import *
from .. import util
from . import generic


if TYPE_CHECKING:
    from ..cluster import Cluster


class Sequence(generic._DynamicObject, generic._CollectionChild):
    """Postgres sequence object. Represents a sequence on a Postgres server.

    Args:
        name (str): name of the sequence
        cluster (str): Postgres cluster object
        schema (str): Schema name
        owner (str): Sequence owner
        oid (int): Sequence OID
        parent (SequenceCollection): parent collection

    Attributes:
        name (str): Sequence name
        cluster (str): Postgres cluster object
        schema (str): Schema name
        owner (str): Sequence owner
        data_type (str): Data type
        start_value (int): Sequence start value
        min_value (int): Sequence minimum value
        max_value (int): Sequence maximum value
        last_value (int): Sequence last value
        increment_by (int): Increment step
        cycle (bool): Whether the sequence is allowed to cycle
        cache_size (int): Cache size
        oid (int): Sequence OID
    """

    def __init__(
        self,
        name: str,
        schema: str = "public",
        owner: str = None,
        cluster: "Cluster" = None,
        parent: "SequenceCollection" = None,
        oid: Optional[int] = None,
    ):
        """Initialize a new Sequence object"""
        super().__init__(kind="SEQUENCE", cluster=cluster, oid=oid, name=name, schema=schema)
        generic._CollectionChild.__init__(self, parent=parent)
        self._owner = owner
        self._schema: str = schema
        self._data_type: Optional[str] = None
        self._start_value: Optional[int] = None
        self._min_value: Optional[int] = None
        self._max_value: Optional[int] = None
        self._increment_by: Optional[int] = None
        self._cycle: Optional[bool] = None
        self._cache_size: Optional[int] = None
        self._last_value: Optional[int] = None

    @property
    def data_type(self):
        return self._data_type

    @data_type.setter
    def data_type(self, value: str):
        if self._data_type != value:
            self._changes["data_type"] = generic._SQLChange(
                obj=self, sql=SQL(f"ALTER SEQUENCE {{fqn}} AS {value}").format(fqn=self._sql_fqn())
            )
            self._data_type = value

    @property
    def start_value(self) -> Optional[int]:
        return self._start_value

    @start_value.setter
    def start_value(self, value: int):
        if self._start_value != value:
            self._changes["start_value"] = generic._SQLChange(
                obj=self,
                sql=SQL("ALTER SEQUENCE {fqn} START WITH {value}").format(
                    fqn=self._sql_fqn(), value=Literal(value)
                ),
            )
            self._start_value = value

    @property
    def min_value(self) -> Optional[int]:
        return self._min_value

    @min_value.setter
    def min_value(self, value: int):
        if self._min_value != value:
            if value == None:
                sql = SQL("ALTER SEQUENCE {fqn} NO MINVALUE").format(fqn=self._sql_fqn())
            else:
                sql = SQL("ALTER SEQUENCE {fqn} MINVALUE {value}").format(
                    fqn=self._sql_fqn(), value=Literal(value)
                )
            self._changes["min_value"] = generic._SQLChange(obj=self, sql=sql)
            self._min_value = value

    @property
    def max_value(self) -> Optional[int]:
        return self._max_value

    @max_value.setter
    def max_value(self, value: int):
        if self._max_value != value:
            if value == None:
                sql = SQL("ALTER SEQUENCE {fqn} NO MAXVALUE").format(fqn=self._sql_fqn())
            else:
                sql = SQL("ALTER SEQUENCE {fqn} MAXVALUE {value}").format(
                    fqn=self._sql_fqn(), value=Literal(value)
                )
            self._changes["max_value"] = generic._SQLChange(obj=self, sql=sql)
            self._max_value = value

    @property
    def increment_by(self) -> Optional[int]:
        return self._increment_by

    @increment_by.setter
    def increment_by(self, value: int):
        if self._increment_by != value:
            self._changes["increment_by"] = generic._SQLChange(
                obj=self,
                sql=SQL("ALTER SEQUENCE {fqn} INCREMENT BY {value}").format(
                    fqn=self._sql_fqn(), value=Literal(value)
                ),
            )
            self._increment_by = value

    @property
    def cycle(self) -> Optional[bool]:
        return self._cycle

    @property
    def cache_size(self) -> Optional[int]:
        return self._cache_size

    @property
    def last_value(self) -> Optional[int]:
        return self._last_value

    @property
    def owner(self) -> Optional[str]:
        return self._owner

    @owner.setter
    def owner(self, owner: str):
        generic._set_ephemeral_attr(self, "owner", owner)

    @property
    def schema(self) -> str:
        return self._schema

    @schema.setter
    def schema(self, schema: str):
        generic._set_ephemeral_attr(self, "schema", schema)

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str):
        generic._set_ephemeral_attr(self, "name", name)

    # methods
    def drop(self, cascade: bool = False):
        """Drops the sequence from the Postgres cluster

        Args:
            cascade (bool): drop dependent objects
        """

        sql = SQL("DROP SEQUENCE {seq}").format(seq=self._sql_fqn())
        if cascade:
            sql += SQL(" CASCADE")
        self.cluster.execute(sql)

    def nextval(self):
        """Retrieve the next value from the sequence"""

        sql = SQL("SELECT nextval(%s)")
        return self.cluster.execute(sql, (self.oid))[0][0]

    def currval(self):
        """Retrieve the next value from the sequence"""

        sql = SQL("SELECT currval(%s)")
        return self.cluster.execute(sql, (self.oid))[0][0]

    def setval(self, value: int, is_called: bool = False):
        """Set sequence's current value and, optionally, is_called flag

        Args:
            is_called (bool): set is_called flag
        """
        if is_called:
            sql = SQL("SELECT setval(%s, %s, true)")
        else:
            sql = SQL("SELECT setval(%s, %s)")
        self.cluster.execute(sql, (self.oid, value))[0][0]

    def refresh(self):
        """Re-initializes the object, refreshing its properties from Postgres cluster"""
        super().refresh()
        if not self._ephemeral:
            sql = util.get_sql("get_sequence") + SQL(" WHERE c.oid = %s")
            result = self.cluster.execute(sql, (self.oid,))
            if not result:
                raise PostgresError("Sequence with oid %s was not found", self.oid)
            mapper = _SequenceMapper(result[0])
            mapper.map(self)


class _SequenceMapper(generic._BaseObjectMapper[Sequence]):
    """Maps out a resultset from a database query to a sequence object"""

    attributes = [
        "name",
        "owner",
        "schema",
        "data_type",
        "start_value",
        "min_value",
        "max_value",
        "increment_by",
        "cycle",
        "cache_size",
        "last_value",
        "oid",
    ]


class SequenceCollection(generic._BaseCollection[Sequence]):
    """An iterable collection of sequences indexed by sequence name.
    For sequences outside of the 'public' schema, index becomes "schemaname.sequencename".
    """

    def __init__(self, cluster: "Cluster"):
        super().__init__(cluster=cluster)
        if cluster:
            self.refresh()

    def refresh(self):
        """Resets any pending changes and refreshes the list of child objects from the cluster"""
        super().refresh()
        sql = util.get_sql("get_sequence")
        result = self.cluster.execute(sql)
        for mapper in [_SequenceMapper(x) for x in result]:
            self[self._index(name=mapper["name"], schema=mapper["schema"])] = mapper.map(
                Sequence(
                    cluster=self.cluster,
                    name=mapper["name"],
                    schema=mapper["schema"],
                    parent=self,
                    oid=mapper["oid"],
                )
            )

    # TODO: uncomment when .create() and .new() are implemented
    # def add(self, sequence: Sequence):
    #     """Adds the database object to the collection, simultaneously creating it on the cluster

    #     Args:
    #         database (Database): initialized database object
    #     """

    #     sequence.cluster = self.cluster
    #     sequence._parent = self
    #     sequence.create()
    #     self[self._index(name=sequence.name, schema=sequence.schema)] = sequence



"""Internal module utilities."""
from typing import Callable, Dict, List, Sequence, Type, TypeVar
from pgmob.sql import SQL
from pathlib import Path
import re
from functools import reduce
from collections import defaultdict
from packaging.version import Version as _Version


_T = TypeVar("_T")


def group_by(key: Callable[..., str], seq: Sequence[_T]) -> Dict[str, List[_T]]:
    """Groups a list by key

    Args:
        key: lambda function to extract the key from a value
        seq: input items

    Returns:
        dict: a dictionary grouped by keys
    """
    return reduce(lambda grp, val: grp[key(val)].append(val) or grp, seq, defaultdict(list))  # type: ignore


class Version(_Version):
    """Version object. Allows to track major, minor, build and revision versions, left to right."""

    @property
    def build(self):
        return self.micro

    @property
    def revision(self):
        return 0 if len(self.release) < 4 else self.release[3]

    def __new__(cls, version):
        try:
            parts = [int(x) for x in version.split(".")]
        except:
            raise ValueError("Unsupported version string. Only dot-separated numbers are supported.")
        if len(parts) == 0:
            raise ValueError("Empty version string")
        return _Version.__new__(cls)


def get_sql(name: str, version: Version = None) -> SQL:
    """Retrieves SQL code from a file in a 'sql' folder

    Args:
        name (str): file name w/o extension"
        version (Version): specific Postgres version to search for

    Returns:
        SQL: sql code
    """
    root_path = Path(__file__).parent / "scripts" / "sql"
    if version:
        filename = f"{name}.sql"
        matcher = re.compile(re.escape(name) + "_(\\d+)\\.sql")
        files = list(root_path.glob(f"{name}_*.sql"))
        files.sort()
        for file in files:
            match = matcher.match(file.name)
            if match and version.major >= int(match[1]):
                filename = file.name
    else:
        filename = f"{name}.sql"
    path = root_path / filename
    with path.open() as sql_file:
        return SQL(sql_file.read())


def get_shell(name: str) -> str:
    """Retrieves shell code from a file in a 'shell' folder

    Args:
        name (str): file name w/o extension

    Returns:
        str: file contents
    """
    path = Path(__file__).parent / "scripts" / f"shell/{name}.sh"
    with path.open() as sql_file:
        return sql_file.read()



from typing import Any, Sequence, Union
from abc import abstractmethod, ABC
from ..sql import Composable


class BaseCursor(ABC):
    """Base Cursor adapter. Abstract class that defines the necessary methods to be overridden by
    the child class that implements Cursor interactions specific to the chosen module.
    """

    # methods that need implementation
    @property
    @abstractmethod
    def statusmessage(self) -> str:
        """Cursor status message"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def closed(self) -> bool:
        """Specifies if the cursor is closed"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def rowcount(self) -> int:
        """Row count of the most recent execute"""
        raise NotImplementedError()

    @abstractmethod
    def close(self) -> None:
        """Close the cursor. The cursor should not be used after this method is called."""
        raise NotImplementedError()

    @abstractmethod
    def execute(self, query: Union[Composable, str], params: tuple = None) -> None:
        """Execute a query with parameters. Should be overridden by the adapter, which should
        provide support for one of the two potential inputs:
            - a query string
            - a Composable object that represent one or more query parts:  SQL, Literal,
              Identifier. Adapter should call the .compose() method of the query object
              to retrieve the parts and reprocess them as needed.

        Args:
            query (Union[Composable, str]): query object or string
            params (tuple): query parameters as a tuple

        Raises:
            NoResultsToFetch: when there are no rows to fetch
            ProgrammingError: when postgres returned an error
            AdapterError: any other Adapter-related error
        """
        raise NotImplementedError()

    @abstractmethod
    def executemany(self, query: Composable, params: Sequence[tuple] = None) -> None:
        """Execute a query with multiple parameter sets. Should be overridden by the adapter, which should
        provide support for all possible Composable objects that represent query parts:
        SQL, Literal, Identifier. Adapter should call the .compose() method of the query object
        to retrieve the parts and reprocess them as needed.
        """
        raise NotImplementedError()

    @abstractmethod
    def mogrify(self, query: Composable, params: tuple = None) -> bytes:
        """Returns a parsed SQL query based on the parameters provided. Should be overridden by the adapter, which should
        provide support for all possible Composable objects that represent query parts:
        SQL, Literal, Identifier. Adapter should call the .compose() method of the query object
        to retrieve the parts and reprocess them as needed.

        Returns:
            bytes: a bytes query string
        """
        raise NotImplementedError()

    @abstractmethod
    def fetchall(self) -> list:
        """Fetch all rows

        Returns:
            list: a ResultSet with the with the query results, if any.

        Raises:
            NoResultsToFetch: when there are no rows to fetch
            ProgrammingError: when postgres returned an error
            AdapterError: any other Adapter-related error
        """
        raise NotImplementedError()

    @abstractmethod
    def fetchone(self) -> Any:
        """Fetch one row

        Returns:
            Any: a single row of the ResultSet

        Raises:
            NoResultsToFetch: when there are no rows to fetch
            ProgrammingError: when postgres returned an error
            AdapterError: any other Adapter-related error
        """
        raise NotImplementedError()

    # default implementations

    def scalar(self, query: Composable, params: tuple = None) -> Any:
        """Same as execute, but returns only the first item of the first row.

        Args:
            query (Any): query object or string
            params (tuple): query parameters as a tuple

        Returns:
            Any: first item of the first row. None if no results have been returned.

        Raises:
            NoResultsToFetch: when there are no rows to fetch
            ProgrammingError: when postgres returned an error
            AdapterError: any other Adapter-related error
        """
        self.execute(query, params)
        result = self.fetchone()
        if len(result) > 0:
            return result[0]
        else:
            return None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class BaseLargeObject(ABC):
    """Base Large Object adapter. Abstract class that defines the necessary methods to be overridden by
    the child class that implements Large Object interactions specific to the chosen module.
    """

    # methods that need implementation
    @property
    @abstractmethod
    def closed(self) -> bool:
        """Specifies if the large object is closed"""
        raise NotImplementedError()

    @abstractmethod
    def close(self):
        """Close and deallocate large object"""
        raise NotImplementedError()

    @abstractmethod
    def truncate(self, length: int = 0):
        """Truncate large object

        Args:
            length (int): size to truncate. 0 to remove
        """
        raise NotImplementedError()

    @abstractmethod
    def write(self, data: bytes) -> int:
        """Write large object

        Returns:
            int: number of bytes written
        """
        raise NotImplementedError()

    @abstractmethod
    def read(self) -> Union[str, bytes]:
        """Read large object

        Returns:
            Union[str, bytes]: string or bytes depending on the large object open mode
        """
        raise NotImplementedError()

    @abstractmethod
    def unlink(self):
        """Remove large object"""
        raise NotImplementedError()

    # default implementations

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class BaseAdapter(ABC):
    """Abstract SQL Provider adapter. Defines required properties and methods to be overriden by adapter implementations."""

    # methods that need implementation
    @abstractmethod
    def cursor(self) -> BaseCursor:
        """Retrieve a cursor object"""
        raise NotImplementedError()

    @abstractmethod
    def lobject(self, oid: int, mode: str) -> BaseLargeObject:
        """Retrieve a large object handler object"""
        raise NotImplementedError()

    @abstractmethod
    def connect(self, *args, **kwargs) -> Any:
        """Retrieve a connection object"""
        raise NotImplementedError()

    @abstractmethod
    def commit(self) -> None:
        """Commit the open transaction"""
        raise NotImplementedError()

    @abstractmethod
    def rollback(self) -> None:
        """Rollback the open transaction"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_connected(self) -> bool:
        """Defines whether the connection is open or closed"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_in_transaction(self) -> bool:
        """Defines whether the transaction is currently open"""
        raise NotImplementedError()

    @abstractmethod
    def close_connection(self) -> None:
        """Close the connection"""
        raise NotImplementedError()

    @abstractmethod
    def get_autocommit(self) -> bool:
        """Get autocommit status"""
        raise NotImplementedError()

    @abstractmethod
    def set_autocommit(self, state: bool) -> None:
        """Set autocommit status"""
        raise NotImplementedError()

    # shared methods

    def get_connection(self) -> Any:
        """Retrieve current connection"""
        return self.connection

    def set_connection(self, connection: Any) -> None:
        """Set current connection"""
        self.connection = connection

    @property
    def has_connection(self) -> bool:
        """Whether the connection is defined"""
        return self.connection is not None



from typing import Any, Callable, Optional, Sequence, Union
import psycopg2  # type: ignore
import psycopg2.sql  # type: ignore
import psycopg2.extras  # type: ignore
import psycopg2.extensions  # type: ignore
from . import ProgrammingError, AdapterError, NoResultsToFetch
from ..sql import SQL, Identifier, Literal, Composable
from .base import BaseAdapter, BaseCursor, BaseLargeObject


class Psycopg2Cursor(BaseCursor):
    """Psycopg2 Cursor Adapter. Defines the cursor behaviour when working with psycopg2 module.

    Args:
        connection (Any): psycopg2 connection object
        *args, **kwargs: any other parameters to be passed on to the cursor
    """

    def __init__(self, connection: Any, *args, **kwargs) -> None:
        self.cursor = connection.cursor(*args, **kwargs)

    @property
    def statusmessage(self) -> str:
        """Cursor status message"""
        return self.cursor.statusmessage

    @property
    def closed(self) -> bool:
        """Specifies if the cursor is closed"""
        return self.cursor.closed

    @property
    def rowcount(self) -> int:
        """Row count of the most recent execute"""
        return self.cursor.rowcount

    def _convert_query(self, query: Union[Composable, str]) -> Union[psycopg2.sql.Composable, str]:
        conv_map = {SQL: psycopg2.sql.SQL, Literal: psycopg2.sql.Literal, Identifier: psycopg2.sql.Identifier}
        if isinstance(query, Composable):
            return psycopg2.sql.Composed([conv_map[part.__class__](part.value()) for part in query.compose()])
        else:
            return query

    def _try_exec(self, func: Callable) -> Any:
        try:
            return func()
        except psycopg2.ProgrammingError as e:
            if str(e) == "no results to fetch":
                raise NoResultsToFetch("No results to fetch") from e
            else:
                raise ProgrammingError("Adapter reported a programming error") from e
        except psycopg2.Error as e:
            raise AdapterError("Adapter reported an error") from e

    def close(self) -> None:
        """Close the currently open cursor"""
        return self.cursor.close()

    def execute(self, query: Union[Composable, str], params: tuple = None) -> None:
        """Execute a query with parameters

        Args:
            query (Union[Composable, str]): query object or string
            params (tuple): query parameters as a tuple

        """
        self._try_exec(lambda: self.cursor.execute(self._convert_query(query), params))

    def executemany(self, query: Union[Composable, str], params: Sequence[tuple] = None) -> None:
        """Execute a query with multiple parameter sets"""
        self._try_exec(lambda: self.cursor.executemany(self._convert_query(query), params))

    def mogrify(self, query: Union[Composable, str], params: tuple = None) -> bytes:
        """Returns a parsed SQL query based on the parameters provided

        Returns:
            bytes: a bytes query string
        """
        return self._try_exec(lambda: self.cursor.mogrify(self._convert_query(query), params))

    def fetchall(self) -> list:
        """Fetch all rows

        Returns:
            list: a ResultSet with the with the query results, if any.
        """
        return self._try_exec(lambda: self.cursor.fetchall())

    def fetchone(self) -> Any:
        """Fetch one row

        Returns:
            Any: a single row of the ResultSet"""
        return self._try_exec(lambda: self.cursor.fetchone())


class Psycopg2LargeObject(BaseLargeObject):
    """Large Object adapter. Implements Large Object interactions when working with psycopg2 module.

    Args:
        connection (Any): connection object
        oid (int): large object oid
        mode (str): connection mode:
            - r  Open for read only
            - w  Open for write only
            - rw Open for read/write
            - n  Don’t open the file
            - b  Return data as bytes
            - t  Decode data as string
    """

    def __init__(self, connection: Any, oid: int, mode: str, *args, **kwargs) -> None:
        self.lobject = connection.lobject(oid, mode=mode, *args, **kwargs)

    @property
    def closed(self) -> bool:
        """Specifies if the large object is closed"""
        return self.lobject.closed

    def close(self):
        """Close and deallocate large object"""
        self.lobject.close()

    def truncate(self, length: int = 0):
        """Truncate large object

        Args:
            length (int): size to truncate. 0 to remove
        """
        self.lobject.truncate(length)

    def write(self, data: bytes) -> int:
        """Write large object

        Returns:
            int: number of bytes written"""
        return self.lobject.write(data)

    def read(self) -> Union[str, bytes]:
        """Read large object

        Returns:
            Union[str, bytes]: string or bytes depending on the large object open mode
        """
        return self.lobject.read()

    def unlink(self):
        """Remove large object"""
        self.lobject.unlink()

    def __del__(self):
        if not self.lobject.closed:
            self.lobject.close()


class Psycopg2Adapter(BaseAdapter):
    """Psycopg2 adapter for PGMob. Implements all necessary protocols to communicate with
    Postgres using psycopg2 module."""

    def __init__(
        self,
        cursor_factory: Optional[psycopg2.extras.DictCursorBase] = None,
    ) -> None:
        self._cursor_factory = cursor_factory
        self.connection: Any = None

    def connect(self, *args, **kwargs) -> Any:
        """Establish connection to the Postgres server."""
        self.connection = psycopg2.connect(*args, **kwargs)

    def cursor(self) -> Psycopg2Cursor:
        """Retrieve the cursor object using the current connection"""
        return Psycopg2Cursor(connection=self.connection, cursor_factory=self._cursor_factory)

    def lobject(self, oid: int, mode: str) -> Psycopg2LargeObject:
        """Retrieve the large object handler using the current connection

        Args:
            oid (int): large object oid
            mode (str): connection mode:
                - r  Open for read only
                - w  Open for write only
                - rw Open for read/write
                - n  Don’t open the file
                - b  Return data as bytes
                - t  Decode data as string
        """
        return Psycopg2LargeObject(connection=self.connection, oid=oid, mode=mode)

    @property
    def is_connected(self) -> bool:
        """Defines whether the connection is open or closed"""
        return bool(self.connection) and not bool(self.connection.closed)

    @property
    def is_in_transaction(self) -> bool:
        """Defines whether the transaction is currently open"""
        return self.connection.status == psycopg2.extensions.STATUS_IN_TRANSACTION

    def commit(self) -> None:
        """Commit the open transaction"""
        self.connection.commit()

    def rollback(self) -> None:
        """Rollback the open transaction"""
        self.connection.rollback()

    def close_connection(self) -> None:
        """Close the connection immediately. The connection will be unusable from this point forward."""
        self.connection.close()

    def get_autocommit(self) -> bool:
        """Get autocommit status"""
        return self.connection.autocommit

    def set_autocommit(self, state: bool) -> None:
        """Set autocommit status"""
        self.connection.autocommit = bool(state)



"""Database objects. Represent databases of the Postgres cluster."""
from typing import TYPE_CHECKING, Any, Dict, Optional, Union
from ..sql import SQL, Composable, Identifier, Literal
from ..errors import *
from .. import util
from . import generic


if TYPE_CHECKING:
    from ..cluster import Cluster


class Database(generic._DynamicObject, generic._CollectionChild):
    """
    Postgres Database object. Represents a database object on a Postgres server.

    Args:
        cluster (str): Postgres cluster object
        name (str): name of the database
        owner (str): Database owner
        encoding (str): Database encoding
        collation (str): Database collation
        is_template (bool): True if the database is a template
        oid (int): Database OID
        parent (DatabaseCollection): parent collection

    Attributes:
        name (str): Database name
        owner (str): Database owner
        encoding (str): Database encoding
        collation (str): Database collation
        character_type (str): Database character type
        is_template (bool): True if the database is a template
        allow_connections (bool): True if the database accepts connections
        connection_limit (int): Database connection limit
        last_sys_oid (int): Last system object ID
        frozen_xid (str): Frozen transaction ID
        min_multixact_id (str): Minimum multi-transaction ID
        tablespace (str): Default tablespace
        acl (str): Database access list
        oid (int): Database OID
    """

    def __init__(
        self,
        name: str,
        cluster: "Cluster" = None,
        parent: "DatabaseCollection" = None,
        owner: str = None,
        encoding: str = None,
        collation: str = None,
        is_template: bool = False,
        oid: Optional[int] = None,
        from_template: str = None,
    ):
        super().__init__(cluster=cluster, name=name, kind="DATABASE", oid=oid)
        generic._CollectionChild.__init__(self, parent=parent)
        self._owner = owner
        self._encoding = encoding
        self._collation = collation
        self._is_template = is_template
        self._from_template = from_template
        self._character_type: Optional[str] = None
        self._allow_connections = True
        self._connection_limit: Optional[int] = None
        self._last_sys_oid: Optional[int] = None
        self._frozen_xid: Optional[int] = None
        self._min_multixact_id: Optional[int] = None
        self._tablespace: Optional[str] = None
        self._acl: Optional[str] = None

    def _modify(self, column: str, value: Any):
        """Internal pg_database modification function"""

        sql = SQL("UPDATE pg_catalog.pg_database SET {column} = {value} WHERE datname = %s").format(
            column=Identifier(column), value=Literal(value)
        )
        self.cluster.execute(sql, self.name)

    # properties
    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str):
        generic._set_ephemeral_attr(self, "name", name)

    @property
    def owner(self) -> Optional[str]:
        return self._owner

    @owner.setter
    def owner(self, owner: str):
        generic._set_ephemeral_attr(self, "owner", owner)

    @property
    def encoding(self) -> Optional[str]:
        return self._encoding

    @property
    def collation(self) -> Optional[str]:
        return self._collation

    @property
    def character_type(self) -> Optional[str]:
        return self._character_type

    @property
    def is_template(self) -> bool:
        return self._is_template

    @is_template.setter
    def is_template(self, value: bool):
        if not isinstance(value, bool):
            raise ValueError("Expected a boolean value")
        if self._is_template != value:
            keyword = "TRUE" if value else "FALSE"
            self._changes["is_template"] = generic._SQLChange(
                obj=self, sql=SQL(f"ALTER DATABASE {{fqn}} IS_TEMPLATE {keyword}").format(fqn=self._sql_fqn())
            )
            self._is_template = value

    @property
    def allow_connections(self) -> bool:
        return self._allow_connections

    @property
    def connection_limit(self) -> Optional[int]:
        return self._connection_limit

    @connection_limit.setter
    def connection_limit(self, value: int):
        if self._connection_limit != value:
            self._changes["connection_limit"] = generic._SQLChange(
                obj=self,
                sql=SQL("ALTER DATABASE {fqn} CONNECTION LIMIT {value}").format(
                    fqn=self._sql_fqn(), value=Literal(value)
                ),
            )
            self._connection_limit = value

    @property
    def last_sys_oid(self) -> Optional[int]:
        return self._last_sys_oid

    @property
    def frozen_xid(self) -> Optional[int]:
        return self._frozen_xid

    @property
    def min_multixact_id(self) -> Optional[int]:
        return self._min_multixact_id

    @property
    def tablespace(self) -> Optional[str]:
        return self._tablespace

    @tablespace.setter
    def tablespace(self, value: str):
        generic._set_ephemeral_attr(self, "tablespace", value)

    @property
    def acl(self) -> Optional[str]:
        return self._acl

    # methods
    def drop(self, terminate_connections: bool = False):
        """Drops the database represented by this object

        Args:
            terminate_connections (bool): Force all connections to be terminated after disabling the database
        """
        try:
            if terminate_connections:
                self.disable()
                self.cluster.terminate(databases=[self.name])
            self.cluster.execute(SQL("DROP DATABASE {database};").format(database=self._sql_fqn()))
        except:
            if terminate_connections:
                self.enable()
            raise

    def refresh(self):
        """Re-initializes the object, refreshing its properties"""
        super().refresh()
        self._from_template = None
        if not self._ephemeral:
            sql = util.get_sql("get_database") + SQL(" WHERE d.oid = %s")
            result = self.cluster.execute(sql, self.oid)
            if not result:
                raise PostgresError("Database not found")
            mapper = _DatabaseMapper(result[0])
            mapper.map(self)

    def disable(self, terminate_connections: bool = False):
        """Disallows connections to a database by modifying pg_database.datallowconn.

        Args:
            terminate_connections (bool): Force all connections to be dropped after disabling the database
        """
        self._modify(column="datallowconn", value=False)
        if terminate_connections:
            self.cluster.terminate(databases=[self.name])

    def enable(self):
        """Enables connections to a database by modifying pg_database.datallowconn"""
        self._modify(column="datallowconn", value=True)

    def create(self):
        """Create a database on the Postgres cluster"""
        self.cluster.execute(self.script(as_composable=True))
        sql = util.get_sql("get_database") + SQL(" WHERE datname = %s")
        _DatabaseMapper(self.cluster.execute(sql, self.name)[0]).map(self)

    def script(self, as_composable: bool = False) -> Union[str, Composable]:
        """Generate a database creation script.

        Args:
            as_composable (bool): return Composable object instead of plain text

        Returns:
            Union[str, Composable]: database creation script
        """
        sql = "CREATE DATABASE {db}"
        params: Dict[str, Composable] = {"db": self._sql_fqn()}
        if self._from_template:
            sql += " TEMPLATE {template}"
            params["template"] = Identifier(self._from_template)
        if self.owner:
            sql += " OWNER {owner}"
            params["owner"] = Identifier(self.owner)
        if self.encoding:
            sql += " ENCODING {encoding}"
            params["encoding"] = Literal(self.encoding)
        if self.collation:
            sql += " LC_COLLATE {locale} LC_CTYPE {locale}"
            params["locale"] = Literal(self.collation)
        if self.is_template:
            sql += " IS_TEMPLATE"

        command = SQL(sql).format(**params)
        if as_composable:
            return command
        else:
            with self.cluster.adapter.cursor() as cur:
                return cur.mogrify(command)


class _DatabaseMapper(generic._BaseObjectMapper[Database]):
    """Maps out a resultset from a database query to a database object"""

    attributes = [
        "name",
        "owner",
        "encoding",
        "collation",
        "character_type",
        "is_template",
        "allow_connections",
        "connection_limit",
        "last_sys_oid",
        "frozen_xid",
        "min_multixact_id",
        "tablespace",
        "acl",
        "oid",
    ]


class DatabaseCollection(generic._BaseCollection[Database]):
    """An iterable collection of databases indexed by database name."""

    def __init__(self, cluster: "Cluster"):
        super().__init__(cluster=cluster)
        if cluster:
            self.refresh()

    def refresh(self):
        """Resets any pending changes and retrieves child objects from the cluster"""
        super().refresh()
        sql = util.get_sql("get_database")
        result = self.cluster.execute(sql)
        for mapper in [_DatabaseMapper(x) for x in result]:
            self[mapper["name"]] = mapper.map(
                Database(cluster=self.cluster, name=mapper["name"], parent=self, oid=mapper["oid"])
            )

    def new(
        self,
        name: str,
        template: str = None,
        owner: str = None,
        is_template: bool = False,
        encoding: str = None,
        collation: str = None,
    ) -> Database:
        """Create a database object on the current Postgres cluster. The object
        is created ephemeral and either needs to be added to the database collection,
        or .create() needs to be executed.

        Args:
            name (str): The name of the database
            template (str): Name of the template database to use
            owner (str): Owner of the database
            encoding (str): Database encoding
            collation (str): Sets LC_COLLATE and LC_CTYPE for the database
            is_template (bool): whether the database is a template
        """
        return Database(
            cluster=self.cluster,
            name=name,
            parent=self,
            from_template=template,
            owner=owner,
            encoding=encoding,
            collation=collation,
            is_template=is_template,
        )

    def add(self, database: Database):
        """Adds the database object to the collection, simultaneously creating it on the cluster

        Args:
            database (Database): initialized database object
        """

        database.cluster = self.cluster
        database._parent = self
        database.create()
        self[database.name] = database



"""Objects in this module are mirrors for remote postgres objects. Any change of the object
(if allowed) will result in affecting that mirrored remote object after executing .apply() method.
When object attributes change, an object schedules a Change closure that would execute once .apply()
is called, which allows users to chain modifications together.

The object state can be refreshed any time by calling .refresh() on the object.
"""

from .databases import Database, DatabaseCollection
from .replication_slots import ReplicationSlot, ReplicationSlotCollection
from .roles import Role, RoleCollection
from .hba_rules import HBARule, HBARuleCollection
from .sequences import Sequence, SequenceCollection
from .tables import Table, TableCollection
from .procedures import (
    Procedure,
    Function,
    Aggregate,
    WindowFunction,
    Volatility,
    ParallelSafety,
    ProcedureVariations,
    ProcedureCollection,
)
from .views import View, ViewCollection
from .schemas import Schema, SchemaCollection
from .large_objects import LargeObject, LargeObjectCollection



"""Adapters are wrappers around SQL Providers, which translate internal module syntax into the proper Provider syntax
"""

from typing import TYPE_CHECKING
from .errors import *


ADAPTERS = [(".psycopg2", "Psycopg2Adapter")]
"""Use the following variable to register adapters for autodiscovery.
Should contain a tuple of module import path: "module.sub" or ".module" - and the Adapter class name.

Example:
    Add a custom adapter ``MyAdapter`` from module ``mymodule``

    >>> import pgmob.adapters
    >>> pgmob.adapters.ADAPTERS.extend(("mymodule", "MyAdapter"))
"""

if TYPE_CHECKING:
    from .base import BaseAdapter


def detect_adapter() -> "BaseAdapter":
    """Detect adapter based on installed modules. You can change the order and add your own
    adapters by modifying the ADAPTERS module variable.

    Returns:
        BaseAdapter: initialized adapter object"""
    from importlib import import_module

    for module, name in ADAPTERS:
        try:
            adapter_module = import_module(module, package="pgmob.adapters")
        except:
            pass
        else:
            return getattr(adapter_module, name)()
    else:
        raise AdapterError(
            "No adapters detected. Try installing one of the following extras: %s", [x[0] for x in ADAPTERS]
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pgmob', 'pgmob.adapters', 'pgmob.objects']

package_data = \
{'': ['*'], 'pgmob': ['scripts/shell/*', 'scripts/sql/*']}

install_requires = \
['packaging>=21.3']

extras_require = \
{'psycopg2': ['psycopg2>=2.8.5,<3'],
 'psycopg2-binary': ['psycopg2-binary>=2.8.5,<3']}

setup_kwargs = {
    'name': 'pgmob',
    'version': '0.1.3a0',
    'description': 'Postgres Managed Objects - a Postgres database management interface',
    'long_description': '![CI](https://github.com/dataplat/pgmob/actions/workflows/CI.yaml/badge.svg)\n# PGMob - PostgreSQL Management Objects\n\nPGMob is a Python package that helps to simplify PostgreSQL administration by providing a layer of abstraction that allows you\nto write simple and easily understandable code instead of having to rely on SQL syntax. It\'s your one tool that helps you to\nmanage your PostgreSQL clusters on any scale and automate routine operations with ease.\n\nPGMob abstracts away the complexity of SQL code and presents a user with a easy to use interface that controls most of\nthe aspects of PostgreSQL administration. It will ensure you won\'t have to switch between Python and SQL while building\nautomation tasks and it will guide you through the process with type helpers and examples.\n\nWith PGMob, you can:\n\n* Control your server while having access to only PostgreSQL protocol\n* Ensure users, databases, and database objects define as you want them\n* Execute backup/restore operations on your server without having to remember the command syntax\n* Script and export your database objects on the fly\n\n\n## Installing\n\nPGMob requires an adapter to talk to PostgreSQL, which it can detect automatically. Currently supported adapters:\n\n* psycopg2\n\nTo install the module without an adapter (you would have to download it by other means) use\n\n```shell\n$ pip install -U pgmob\n```\n\nTo include the adapter, use pip extras feature:\n\n```shell\n$ pip install -U pgmob[psycopg2]\n```\n\n## Documentation\n\nTBD\n\n## Example code\n\n```python\nfrom pgmob import Cluster\n\ncluster = Cluster(host="127.0.0.1", user="postgres", password="s3cur3p@ss")\n\n# Execute a simple query with parameters\ncluster.execute("SELECT tableowner FROM pg_tables WHERE tablename LIKE %s", "pg*")\n\n# Create a new database owner and reassign ownership\nowner_role = cluster.roles.new(name="db1owner", password="foobar")\nowner_role.create()\ndb = cluster.databases["db1"]\ndb.owner = owner_role.name\ndb.alter()\n\n# Modify pg_hba on the fly:\nentry = "host all all 127.0.0.1/32 trust"\nif entry not in cluster.hba_rules:\n    cluster.hba_rules.extend(entry)\n    cluster.hba_rules.alter()\n\n# clone an existing role\nsql = cluster.roles["somerole"].script()\ncluster.execute(sql.replace("somerole", "newrole"))\n\n# control access to your database\ncluster.terminate(databases=["somedb"], roles=["someapp"])\ncluster.databases["someotherdb"].disable()\n\n# run backups/restores\nfrom pgmob.backup import FileBackup, FileRestore\n\nfile_backup = FileBackup(cluster=cluster)\nfile_backup.options.schema_only = True\nfile_backup.backup(database="db1", path="/tmp/db.bak")\n\ncluster.databases.new("db2").create()\nfile_restore = FileRestore(cluster=cluster)\nfile_restore.restore(database="db2", path="/tmp/db.bak")\n\n# create, modify, and drop objects\ncluster.schemas.new("app_schema").create()\nfor t in [t for t in cluster.tables if t.schema == "old_schema"]:\n    t.schema = "app_schema"\n    t.alter()\ncluster.schemas["old_schema"].drop()\n```\n\n## Dynamic objects and collections\n\nEach Python object in PGMob is asynchronously connected to the corresponding object on the server. When changing object attributes,\none only changes the local object. In order to push the changes to the server, one needs to execute the `.alter()` method of the dynamic\nobject solidyfing the changes on the server.\n\nWhen working with collections, such as tables, procedures, and others, you can retrieve corresponding objects using their name as index:\n\n```python\ncluster.tables["tablename"]\n# or, in case the schema is not public\ncluster.tables["myschema.tablename"]\n```\n\nHowever, you can iterate upon such collections as if they were a list:\n\n```python\nfor t in cluster.tables:\n    t.owner = "new_owner"\n    t.alter()\nif "myschema.tab1" in cluster.tables:\n    cluster.tables["myschema.tab1"].drop()\n```\n\nThis helps the developer to write a concise and readable code when working with PostgreSQL objects.\n\n## Links\n\nTBD\n',
    'author': 'Kirill Kravtsov',
    'author_email': 'nvarscar@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/dataplat/pgmob/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4',
}


setup(**setup_kwargs)



"""A list of PGMob errors raised by the module"""
import logging


class PostgresError(Exception):
    """An error raised by the PGMob module"""

    def __init__(self, *args):
        super().__init__(*args)
        logger = logging.getLogger(__name__)
        self._arguments = None
        self._message = "PostgresError has been raised"
        if args:
            self._message = args[0]
            if len(args) > 1:
                self._arguments = args[1:]
        if self._arguments:
            logger.debug(self._message, self._arguments)
        else:
            logger.debug(self._message)

    def __str__(self):
        return self._message % self._arguments if self._arguments else self._message


class PostgresShellCommandError(PostgresError):
    """An error raised when OS Command has returned a non-zero exit code"""

    def __init__(self, command, error_message):
        message = f"Error while executing shell command [{command}]: %s"
        super().__init__(message, error_message)


class PostgresTypeError(TypeError):
    """Incorrect type provided"""

    pass


class PostgresNameError(NameError):
    """Incorrect name provided"""

    pass
